"""Utilities for coil intensity non-unformity correction (aka bias correction).
"""

import numpy as np
from mrrt.utils import get_array_module


__all__ = ["compute_bias_field"]


def compute_bias_field(
    recon, affine=None, down=None, basename="xtrue", xp=None
):
    # TODO: call N4BiasFieldCorrection via SimpleITK or the ITK Python wrappers
    #       instead of via shell
    import nibabel as nib
    import subprocess

    xp, on_gpu = get_array_module(recon, xp)
    if affine is None:
        affine = np.eye(4)
    ndim = recon.ndim
    recon_abs = np.abs(recon)
    recon_abs = (32767 * recon_abs / recon_abs.max()).astype(np.uint16)
    if xp != np:
        recon_abs = recon_abs.get()
    recon_nii = nib.Nifti1Image(recon_abs, affine=affine)
    recon_nii.to_filename("{0}.nii".format(basename))

    if down is None:
        if ndim == 3:
            down = 3
            cstr = "[200x200x200x200]"
        else:
            down = 1
            cstr = "[400x200x200]"
    cmd = (
        "N4BiasFieldCorrection"
        " -d {0}"
        " -s {1} -c " + cstr + " -i {2}.nii"
        " -o [{2}_biascor.nii, {2}_biasfield.nii]"
    )
    cmd = cmd.format(ndim, down, basename)
    try:
        subprocess.check_output(cmd, shell=True)
    except RuntimeError:
        print(
            "Subprocess failed: is ANTS N4BiasFieldCorrection "
            "on the system path?"
        )
        raise
    bias = nib.load("{0}_biasfield.nii".format(basename)).get_data()
    np.save("{}_biasfield".format(basename), np.asarray(bias, dtype=np.float32))
    return xp.asarray(bias)



from itertools import product

import numpy as np
import pytest

from mrrt.mri.sim import mri_object_2d, mri_object_3d
from mrrt.mri import mri_partial_fourier_nd
from mrrt.utils import config, ifftnc, ImageGeometry

all_xp = [np]
if config.have_cupy:
    import cupy

    all_xp += [cupy]


# TODO: add test case with non-zero phase


@pytest.mark.parametrize(
    "xp, dtype", product(all_xp, [np.complex64, np.complex128])
)
def test_partial_fourier_2d(xp, dtype):
    shape = (256, 192)
    ig = ImageGeometry(shape, distances=(1, 1), offsets="dsp")
    obj = mri_object_2d(ig.fov)
    coords = ig.fgrid()

    # fully-sampled k-space
    kspace_full = xp.asarray(obj.kspace(*coords), dtype=dtype)

    # partial-Fourier k-space
    pf_fractions = (0.6, 0.7)
    nkeep = [int(f * s) for f, s in zip(pf_fractions, shape)]
    pf_mask = xp.zeros(shape, dtype=xp.bool)
    pf_mask[: nkeep[0], : nkeep[1]] = 1
    kspace_pf = kspace_full[: nkeep[0], : nkeep[1]]

    # direct reconstruction using zero-filled k-space
    direct_recon = ifftnc(kspace_full * pf_mask)

    # partial Fourier reconstruction
    pf_recon = mri_partial_fourier_nd(kspace_pf, pf_mask)
    # dtype is preserved
    assert pf_recon.dtype == dtype

    # ground truth image
    # x_true = xp.asarray(obj.image(*ig.grid()))

    # recon from fully sampled k-space
    x_full = xp.asarray(ifftnc(kspace_full))

    # Error of partial-Fourier recon should be much less than for zero-filling
    mse_pf = xp.mean(xp.abs(x_full - pf_recon) ** 2)
    mse_direct = xp.mean(xp.abs(x_full - direct_recon) ** 2)
    assert mse_pf < 0.2 * mse_direct


@pytest.mark.parametrize(
    "xp, dtype, pf_fractions",
    product(
        all_xp,
        [np.complex64, np.complex128],
        [(1, 0.6, 1), (0.6, 0.7, 1), (0.65, 0.65, 0.65)],
    ),
)
def test_partial_fourier_3d(xp, dtype, pf_fractions):
    shape = (128, 128, 64)
    ig = ImageGeometry(shape, distances=(1,) * len(shape), offsets="dsp")
    obj = mri_object_3d(ig.fov)
    coords = ig.fgrid()

    # fully-sampled k-space
    kspace_full = xp.asarray(obj.kspace(*coords), dtype=dtype)

    # partial-Fourier k-space
    nkeep = [int(f * s) for f, s in zip(pf_fractions, shape)]
    pf_mask = xp.zeros(shape, dtype=xp.bool)
    pf_mask[: nkeep[0], : nkeep[1]] = 1
    kspace_pf = kspace_full[: nkeep[0], : nkeep[1]]

    # direct reconstruction using zero-filled k-space
    direct_recon = ifftnc(kspace_full * pf_mask)

    # partial Fourier reconstruction
    pf_recon = mri_partial_fourier_nd(kspace_pf, pf_mask)
    # dtype is preserved
    assert pf_recon.dtype == dtype

    # ground truth image
    # x_true = xp.asarray(obj.image(*ig.grid()))

    # recon from fully sampled k-space
    x_full = xp.asarray(ifftnc(kspace_full))

    # Error of partial-Fourier recon should be much less than for zero-filling
    mse_pf = xp.mean(xp.abs(x_full - pf_recon) ** 2)
    mse_direct = xp.mean(xp.abs(x_full - direct_recon) ** 2)
    assert mse_pf < 0.25 * mse_direct



import numpy as np
from scipy.special import ellipk, ellipe

from mrrt.utils import get_array_module  # TODO: move

__all__ = ["sensemap_sim"]


def _mri_smap1(x, y, z, a, xp=None):
    """circular coil in "x-y plane" of radius a

    note that coil x-y plane is not same as object x-y plane!
    """
    xp, on_gpu = get_array_module(x)
    x = x / a
    y = y / a
    z = z / a
    r = xp.sqrt(x * x + y * y)
    r[r == 0] = 1e-7  # avoid divide by zero

    zsq = z * z
    rp1sq = (r + 1) ** 2

    M = 4 * r / (rp1sq + zsq)
    # elliptic integral of the first and second kind
    if xp != np:
        # CuPy doesn't currently have ellipk or ellipe so have to transfer to
        # the CPU to evaluate the elliptic integrals.
        M = M.get()
    (K, E) = ellipk(M), ellipe(M)
    if xp != np:
        K, E = map(xp.asarray, (K, E))

    # B_z in eqn (18) in grivich:00:tmf
    tmp = (rp1sq + zsq) ** (-0.5)
    tmp2 = (1 - r) ** 2 + zsq
    rsq = r * r
    smap_z = 2 * tmp * (K + (1 - rsq - zsq) / tmp2 * E)
    smap_z /= a

    # B_r in eqn (17) in grivich:00:tmf
    smap_r = 2 * z / r * tmp * ((1 + rsq + zsq) / tmp2 * E - K)

    bad = xp.abs(r) < 1e-6
    smap_r[bad] = 3 * xp.pi * z[bad] / ((1 + z[bad] ** 2) ** 2.5) * r[bad]
    smap_r /= a

    if xp.any(xp.isnan(smap_r)) or xp.any(xp.isnan(smap_z)):
        raise Exception("Nan found in smap")

    phi = xp.arctan2(y, x)
    smap_x = smap_r * xp.cos(phi)
    smap_y = smap_r * xp.sin(phi)

    return (smap_x, smap_y, smap_z)


def sensemap_sim(
    shape=(64, 64),
    spacings=(3, 3),
    ncoil=8,
    rcoil=100,
    orbit=360,
    orbit_start=None,
    coil_distance=1.5,
    nring=1,
    dz_coil=None,
    scale="default",
    dtype=np.complex128,
    xp=np,
):
    """Simulate sensitivity maps for sensitivity-encoded MRI.

    Parameters
    ----------
    shape : tuple of int
        The image or volume size. Must have either two or 3 elements.
    spacings : tuple of int
        The voxel dimensions. Should have equal length to shape. If an integer
        is provided, the same value is assumed for all axes.
    ncoil : int, optional
        Total number of coils (for all rings).
    rcoil : int or None optional
        The radius of an individual coil element. If None, it will default to
        ``shape[0] * spacings[0] / 4``.
    coil_distance : float, optional
        Distance of coil center from isocenter for central ring of coils as a
        multiple of fov_x, where ``fov_x = nx * dx``.
    orbit : float, optional
        Angular range around the cylinder covered.
    orbit_start : list of float, optional
        Can be a list of length nring. Offsets to the start of each ring in
        degrees.
    nring : int, optional
        Number of rings of coils (along a cylinder in z).
    dz_coil : float, optional
        Ring spacing in z (defaults to ``shape[2] * spacings[2] / nring``).
    scale : {'default', 'ssos'}, optional
        If ``'ssos'``, scale so that the sqrt of the sum of square of the
        center equals 1.
    dtype : {np.complex64, np.complex128}, optional
        Data type for the generated sensitivity maps.

    Returns
    -------
    smap : array
        The simulated sensitivity maps. ``smap.shape`` will be
        ``shape + (ncoils,)``.

    Notes
    -----
    based on [1]_
    Adapted from a Matlab implementation Copyright 2005-2016,
    Jeff Fessler, Amanda Funai and Mei Le, University of Michigan
    Python adaptation by Gregory Lee.

    References
    ----------
    .. [1] Grivich MI, Jackson DP.  The magnetic field of current-carrying
    polygons: An application of vector field rotations.
    Am. J. Phys. 68, 469 (2000).  doi:10.1119/1.19461
    """
    if len(shape) == 2:
        shape = tuple(shape) + (1,)
        if len(spacings) == 2:
            spacings = tuple(spacings) + (1,)

    if len(shape) != 3:
        raise ValueError("shape must have length 2 or 3")
    if np.isscalar(spacings):
        spacings = (spacings,) * 3
    nx, ny, nz = shape
    dx, dy, dz = spacings
    if rcoil is None:
        rcoil = dx * nx / 2 * 0.50
    if dz_coil is None:
        dz_coil = dz * nz / nring

    coils_per_ring = int(np.round(ncoil / nring))
    if nring * coils_per_ring != ncoil:
        raise ValueError("nring must be a divisor of ncoil")

    if dtype == np.complex128:
        real_dtype = np.float64
    elif dtype == np.complex64:
        real_dtype = np.float32
    else:
        raise ValueError("unsupported dtype")

    # coil radii
    rlist = rcoil * xp.ones((coils_per_ring, nring), dtype=real_dtype)

    # position of coil center (x, y, z)
    plist = xp.zeros((coils_per_ring, nring, 3), dtype=real_dtype)
    # normal vector (inward) from coil center
    nlist = xp.zeros((coils_per_ring, nring, 3), dtype=real_dtype)
    # unit vector orthogonal to normal vector in x-y
    olist = xp.zeros((coils_per_ring, nring, 3), dtype=real_dtype)
    # upward vector
    # ulist = xp.zeros((coils_per_ring, nring, 3), dtype=real_dtype)

    if orbit_start is None:
        orbit_start = [0] * nring
    if xp.isscalar(orbit_start):
        orbit_start = orbit_start[0] * nring
    elif len(orbit_start) == 1:
        orbit_start = [orbit_start[0]] * nring
    elif len(orbit_start) != nring:
        raise ValueError(
            "orbit_start should be a single value or a list of length nring"
        )

    # circular coil configuration, like head coils
    # list of angles in radians
    # ncoil) / coils_per_ring
    alist = xp.deg2rad(orbit) * xp.linspace(0, 1, coils_per_ring + 1)[:-1]
    z_ring = (xp.arange(1, nring + 1) - (nring + 1) / 2) * dz_coil
    for ir in range(nring):
        alist_ring = alist + xp.deg2rad(orbit_start[ir])
        for ic in range(coils_per_ring):
            phi = alist_ring[ic]
            maxval = xp.max(xp.asarray([nx / 2 * dx, ny / 2 * dy]))
            rad = maxval * coil_distance
            cp = xp.cos(phi)
            sp = xp.sin(phi)
            plist[ic, ir, :] = xp.array(
                [rad * cp, rad * sp, z_ring[ir]], dtype=real_dtype
            )
            nlist[ic, ir, :] = -xp.array(
                [cp, sp, 0], dtype=real_dtype  # cylinder
            )
            olist[ic, ir, :] = xp.array([-sp, cp, 0], dtype=real_dtype)
            # ulist[ic, ir, :] = xp.array([0, 0, 1], dtype=real_dtype)

    # object coordinates for slice z=0
    x = (xp.arange(1, nx + 1, dtype=real_dtype) - (nx + 1) / 2) * dx
    y = (xp.arange(1, ny + 1, dtype=real_dtype) - (ny + 1) / 2) * dy
    z = (xp.arange(1, nz + 1, dtype=real_dtype) - (nz + 1) / 2) * dz
    xx, yy, zz = xp.meshgrid(x, y, z, indexing="ij")
    # zz = xp.zeros_like(xx)

    smap = xp.zeros((nx, ny, nz, coils_per_ring, nring), dtype=dtype)
    for ir in range(nring):
        for ic in range(coils_per_ring):

            if nlist[ic, ir, 2] or olist[ic, ir, 2]:
                # assume z component of plist and nlist are 0
                raise Exception("Unsupported")

            # rotate coordinates to correspond to coil orientation
            zr = (
                (xx - plist[ic, ir, 0]) * nlist[ic, ir, 0]
                + (yy - plist[ic, ir, 1]) * nlist[ic, ir, 1]
                + (zz - plist[ic, ir, 2]) * nlist[ic, ir, 2]
            )
            xr = xx * nlist[ic, ir, 1] - yy * nlist[ic, ir, 0]
            yr = zz - plist[ic, ir, 2]  # translate along object z axis

            # compute sensitivity vectors in coil coordinates
            (sx, sy, sz) = _mri_smap1(xr, yr, zr, rlist[ic, ir])

            # coil response depends on tranverse magnetization only?
            # TODO: unsure if this should depend on sy and ulist in 3D
            bx = sz * nlist[ic, ir, 0] + sx * olist[ic, ir, 0]
            by = sz * nlist[ic, ir, 1] + sx * olist[ic, ir, 1]
            # bz = sz * nlist[ic, ir, 2] + sx * olist[ic, ir, 2]
            smap[:, :, :, ic, ir] = bx + 1.0j * by

    smap = smap * rlist[0] / (2 * xp.pi)  # trick: scale so near unity maximum

    if nz == 1:
        smap = smap.reshape((nx, ny, ncoil), order="F")
        scale_center = 1 / xp.sqrt(
            xp.sum(xp.abs(smap[nx // 2, ny // 2, :] ** 2))
        )
    else:
        smap = smap.reshape((nx, ny, nz, ncoil), order="F")
        scale_center = 1 / xp.sqrt(
            xp.sum(xp.abs(smap[nx // 2, ny // 2, nz // 2, :] ** 2))
        )

    if scale.lower() == "ssos":
        smap *= scale_center
    elif scale != "default":
        raise ValueError("unrecognized scale: {}".format(scale))

    return smap



import numpy as np
import pytest

from mrrt.mri.sim import mri_object_1d, mri_object_2d, mri_object_3d

from mrrt.utils import ImageGeometry, ifftnc


# TODO: add test for mri_object_2d_multispectral


@pytest.mark.parametrize("dtype", [np.float32, np.float64])
def test_mri_object_1d(dtype):
    obj = mri_object_1d(fov=(240,), units="mm", dtype=dtype)
    ig = ImageGeometry((240,), distances=(1.0,))

    img = obj.image(*ig.grid())
    ksp = obj.kspace(*ig.fgrid()) / ig.distances[0]

    norm_err = np.linalg.norm(img - ifftnc(ksp)) / np.linalg.norm(img)
    assert norm_err < 0.1


@pytest.mark.parametrize("dtype", [np.float32, np.float64])
def test_mri_object_2d(dtype):
    obj = mri_object_2d(fov=(240, 240), units="mm", dtype=dtype)
    ig = ImageGeometry((240, 240), distances=(1.0, 1.0))

    img = obj.image(*ig.grid())
    ksp = obj.kspace(*ig.fgrid()) / np.prod(ig.distances)

    norm_err = np.linalg.norm(img - ifftnc(ksp)) / np.linalg.norm(img)
    assert norm_err < 0.1


@pytest.mark.parametrize("dtype", [np.float32, np.float64])
def test_mri_object_3d(dtype):
    obj = mri_object_3d(fov=(120, 120, 40), units="mm", dtype=dtype)
    ig = ImageGeometry((240, 240, 80), distances=(0.5, 0.5, 0.5))

    img = obj.image(*ig.grid())
    ksp = obj.kspace(*ig.fgrid()) / np.prod(ig.distances)

    norm_err = np.linalg.norm(img - ifftnc(ksp)) / np.linalg.norm(img)
    assert norm_err < 0.2



from __future__ import division, print_function, absolute_import
import numpy as np

from mrrt.utils import get_array_module  # TODO: move

__all__ = ["mri_exp_mult"]


def mri_exp_mult(A, u, v, xp=None, debug=False):
    """
    Y = A.H * exp(-u * v.T)      [L x M]
    """
    xp, on_gpu = get_array_module(A, xp)

    if debug:
        print("A.shape = {}".format(A.shape))
        print("u.shape = {}".format(u.shape))
        print("v.shape = {}".format(v.shape))

    if A.ndim == 1:
        A = A[:, np.newaxis]
    elif A.ndim != 2:
        raise ValueError("A must be 2d")
    n, segs = A.shape

    if u.ndim > 1 or v.ndim > 1:
        raise ValueError("u, v must be 1d")

    u = u[:, np.newaxis]
    v = v[np.newaxis, :]
    m = v.size

    if n != u.size:
        raise ValueError(
            "Inconsistent Dimensions: n={}, u.shape[1]={}".format(n, u.shape[1])
        )

    if debug:
        print("mri_exp_mult:  n={}, m={}, segs={}".format(n, m, segs))

    if v.size < 4e6:
        tmp = -xp.dot(u, v)
        xp.exp(tmp, out=tmp)
        return xp.dot(xp.conj(A.T), tmp)
    else:
        # break into chunks to reduce memory required
        nchunks = int(xp.ceil(v.size / 1e6))
        nper = int(xp.ceil(v.size / nchunks))
        arrays = []
        for ci in range(nchunks):
            print("computing chunk {} of {}".format(ci + 1, nchunks))
            if ci == nchunks - 1:
                sl = (slice(None), slice(ci * nper, v.size))
            else:
                sl = (slice(None), slice(ci * nper, (ci + 1) * nper))
            tmp = -xp.dot(u, v[sl])
            tmp = xp.exp(tmp, out=tmp)
            arrays.append(xp.dot(xp.conj(A.T), tmp))
        return xp.concatenate(arrays, axis=1)



""" Basic benchmarking of some Non-Cartesian MRI reconstruction cases. """
from __future__ import print_function, division

import numpy as np

from matplotlib import pyplot as plt

from mrrt.mri.operators.tests.test_mri_noncartesian import _test_mri_multi


def _split_timings(bench_tuple, print_timings=False):
    t_adj = [(k, v["MRI: adjoint"]) for k, v in bench_tuple.items()]
    adjoint_times = sorted(t_adj, key=lambda x: x[1], reverse=True)

    t_for = [(k, v["MRI: forward"]) for k, v in bench_tuple.items()]
    forward_times = sorted(t_for, key=lambda x: x[1], reverse=True)

    t_create = [(k, v["MRI: object creation"]) for k, v in bench_tuple.items()]
    creation_times = sorted(t_create, key=lambda x: x[1], reverse=True)

    t_norm = [(k, v["MRI: norm"]) for k, v in bench_tuple.items()]
    norm_times = sorted(t_norm, key=lambda x: x[1], reverse=True)

    if print_timings:
        print("creation times")
        for k, v in creation_times[::-1]:
            print("\t{}: {}".format(k, v))
        print("forward times")
        for k, v in forward_times[::-1]:
            print("\t{}: {}".format(k, v))
        print("adjoint times")
        for k, v in adjoint_times[::-1]:
            print("\t{}: {}".format(k, v))
        print("norm times")
        for k, v in norm_times[::-1]:
            print("\t{}: {}".format(k, v))
    return creation_times, forward_times, adjoint_times, norm_times


def _autolabel(ax, rects, fmt="%0.3g"):
    """
    Attach a text label to the right of each bar displaying its width
    """
    # use mean_width to determine text spacing so that it will be the same
    # for all bars
    mean_width = np.mean([rect.get_width() for rect in rects])
    max_width = 0
    for rect in rects:
        width = rect.get_width()
        t = ax.text(
            width + 0.02 * mean_width,
            rect.get_y() + rect.get_height() / 2.0,
            fmt % width,
            va="center",
            ha="left",
        )
        if width > max_width:
            t_max = t
            max_width = width

    # determine size of the text as in:
    #   http://stackoverflow.com/questions/5320205/matplotlib-text-dimensions
    bb = t_max.get_window_extent(renderer=ax.get_figure().canvas.get_renderer())
    bb_ax = ax.get_window_extent()
    stretch_factor = (bb_ax.width + bb.width) / bb_ax.width
    # stretch xlim a bit to accomodate the labels
    ax.set_xlim((ax.get_xlim()[0], stretch_factor * ax.get_xlim()[1]))


def _plot_timings(
    labels_times_tuple, title=None, xlabel=None, xlim=None, ax=None, fmt="%0.3g"
):
    if ax is None:
        fig, ax = plt.subplots(1, 1, figsize=[12, 6])
    nlabel = len(labels_times_tuple)
    labels = [a[0] for a in labels_times_tuple]
    labels = [l.replace("single", "s") for l in labels]
    labels = [l.replace("double", "d") for l in labels]
    labels = [l.replace("complex", "C") for l in labels]
    labels = [l.replace("real", "R") for l in labels]
    rects = ax.bar(
        np.zeros(nlabel),
        height=0.8,
        bottom=np.arange(nlabel),
        width=[a[1] for a in labels_times_tuple],
        tick_label=labels,
        orientation="horizontal",
        align="center",
    )
    _autolabel(ax, rects, fmt=fmt)
    if title is not None:
        ax.set_title(title)
    if xlabel is not None:
        ax.set_xlabel(xlabel)
    # ax.axis('tight')
    if xlim is not None:
        ax.set_xlim(xlim)
    #        mean_width = np.mean([r.get_width() for r in rects])
    #        for rect in rects:
    #            width = rect.get_width()
    #            if width < xlim[-1]:
    #                t = ax.text(xlim[-1] + 0.02*mean_width,
    #                            rect.get_y() + rect.get_height()/2.,
    #                            '%0.3g' % width,
    #                            va='center', ha='left')

    # center the text of any multi-line labels
    ax.yaxis.set_ticklabels(ax.yaxis.get_ticklabels(), multialignment="center")
    return ax


if False:
    # BART CPU without Toeplitz 2873 ms
    # BART CPU with Toeplitz 2154 ms
    # BART GPU with Toeplitz 2027 ms
    # BART GPU without Toeplitz 2027 msG
    # ('BART (CPU)', 2154),

    from matplotlib import rcParams

    rcParams["font.size"] = 14

    labels_times_tuple = (
        ("Michigan\nIRT", 5861),
        # fftn/ifftn with FFTW_MEAURE planner effort
        ("PythonCPU\n(16 cores)", 2670),
        ("BART (GPU)", 2027),
        ("gpuNUFFT", 353),
        ("PythonGPU\n(No Precomp)", 58.5),
        ("PythonGPU\n(Full_Precomp)", 56.2),
    )
    fig1, ax1 = plt.subplots(1, 1)
    _plot_timings(
        labels_times_tuple[::-1],
        title="96x96x96 with 16 coils",
        xlabel="$A^{H}A$ execution time (ms)",
        ax=ax1,
        fmt="%d",
    )
    plt.tight_layout()

    labels_times_tuple = (
        ("Single Coil", 122),
        ("Single Coil\n(no Precomp)", 138),
        ("16 Coils", 588),
        ("16 Coils\n(no Precomp)", 727),
        ("16 Coils +\nFieldmap\n(6 time segments)", 3090),
        ("16 Coils +\nFieldmap\n(6 time segments)\n(No Precomp)", 4080),
    )
    fig2, ax2 = plt.subplots(1, 1)
    _plot_timings(
        labels_times_tuple[::-1],
        title="PythonGPU\n(256x256x256 voxels)",
        xlabel="$A^{H}A$ execution time (ms)",
        ax=ax2,
        fmt="%d",
    )
    plt.tight_layout()

    labels_times_tuple = (
        ("SimpleITK\nCPU\n(16 cores))", 706),
        ("PythonGPU", 57.7),
        ("PythonGPU\n(warp field\npreloaded)", 37.8),
    )
    fig3, ax3 = plt.subplots(1, 1)
    _plot_timings(
        labels_times_tuple[::-1],
        title=("Warp with cubic interpolation" "\n(200x148x96 complex array)"),
        xlabel="execution time (ms)",
        ax=ax3,
        fmt="%d",
    )
    plt.tight_layout()

    out_dir = "/home/lee8rx/Dropbox/Grants/SSE/LaTex_NIH/Figures/"
    from os.path import join as pjoin

    fig1.savefig(pjoin(out_dir, "comparison2.png"))
    fig1.savefig(pjoin(out_dir, "comparison2.eps"))
    fig1.savefig(pjoin(out_dir, "comparison2.pdf"))
    fig2.savefig(pjoin(out_dir, "gpu_cases.png"))
    fig2.savefig(pjoin(out_dir, "gpu_cases.eps"))
    fig2.savefig(pjoin(out_dir, "gpu_cases.pdf"))
    fig3.savefig(pjoin(out_dir, "warp.png"))
    fig3.savefig(pjoin(out_dir, "warp.eps"))
    fig3.savefig(pjoin(out_dir, "warp.pdf"))
    # 96x96x96x16coils with 400 shots instead of 150
    # PythonGPU: 90.1 ms
    # BART CPU with Toeplitz: 2205 ms
    # BART GPU with Toeplitz: 2223 ms

    # for data of size 128x128x128
    ("ShearLab3D (dec)", 14.626),
    # ('ShearLab3D (thresh)', 5.264),
    ("ShearLab3D (rec)", 16.303),
    ("UDCT (dec)", 3.37),
    ("UDCT (rec)", 1.71),
    ("DAST (Matlab) (dec)", 1.298),  # (double precision)
    # ('DAST (Matlab) (thresh)', 8.145),
    ("DAST (Matlab) (rec)", 1.577),
    ("DAST (Python) (dec)", 0.845),  # (single precision)
    # ('DAST (Python) (dec)', 0.786),  # w/ batchFFT
    # ('DAST (Python) (thresh)', 1.668),
    ("DAST (Python) (rec)", 0.865),
    ("Surfacelet (dec, R=6.4, mex)", 1.538)
    ("Surfacelet (dec, R=6.4, pure Matlab)", 2.137)
    ("Surfacelet (rec, R=6.4, mex)", 1.334)
    ("Surfacelet (rec, R=6.4, pure Matlab)", 2.282)
    # ('DAST (Python) (rec)', 0.770),   # w/ batchFFT


def _plot_all_timings(
    creation_times,
    forward_times,
    adjoint_times,
    norm_times=None,
    title_substring="",
    uniform_xscale=False,
    plot_creation_times=True,
):
    nsub = 2
    if plot_creation_times:
        nsub += 1
    if norm_times is not None:
        nsub += 1
    if nsub == 4:
        fig, axes = plt.subplots(2, 2, figsize=[16, 16])
        axes = axes.ravel()
    else:
        fig, axes = plt.subplots(nsub, 1, figsize=[16, 12])
    max_time_create = np.max([t[1] for t in creation_times])
    max_time_forward = np.max([t[1] for t in forward_times])
    max_time_adjoint = np.max([t[1] for t in adjoint_times])
    max_time = np.maximum(max_time_forward, max_time_adjoint)
    if norm_times is not None:
        max_time_norm = np.max([t[1] for t in norm_times])
        max_time = np.maximum(max_time, max_time_norm)
    if uniform_xscale:
        max_time = np.maximum(max_time, max_time_create)
        xlim1 = [0, max_time]
        xlim = xlim1
    else:
        xlim1 = None
        xlim = None  # [0, max_time]

    print([0, max_time])
    n = 0
    if plot_creation_times:
        _plot_timings(
            creation_times,
            title="creation: " + title_substring,
            xlabel="time (s)",
            xlim=xlim1,
            ax=axes[n],
        )
        n += 1
    _plot_timings(
        forward_times,
        title="forward: " + title_substring,
        xlabel="time (s)",
        xlim=xlim,
        ax=axes[n],
    )
    n += 1
    _plot_timings(
        adjoint_times,
        title="adjoint: " + title_substring,
        xlabel="time (s)",
        xlim=xlim,
        ax=axes[n],
    )
    n += 1
    if norm_times is not None:
        _plot_timings(
            norm_times,
            title="norm: " + title_substring,
            xlabel="time (s)",
            xlim=xlim,
            ax=axes[n],
        )
        n += 1
    if uniform_xscale and plot_creation_times:
        axes[0].set_xticks([])
        axes[1].set_xticks([])
        axes[0].set_xlabel("")
        axes[1].set_xlabel("")
    if not plot_creation_times:
        axes[0].set_xticks([])
        axes[0].set_xlabel("")

    return axes


def _save_bench(save_dir, fname_base, timing_tuples, axes=None):
    import pickle
    from os.path import join as pjoin

    if axes is not None:
        axes[0].get_figure().savefig(pjoin(save_dir, fname_base + ".png"))
    with open(pjoin(save_dir, fname_base + ".p"), "wb") as f:
        pickle.dump(timing_tuples, f)


def _old_bar_graph(bench_tuple):
    t = bench_tuple
    fields = ["MRI: adjoint", "MRI: forward", "MRI: object creation"]
    for field in fields:
        sorted_keys = sorted(t.keys())
        plt.figure()
        y = [t[key][field] for key in sorted_keys]
        x = np.arange(len(y))
        width = 0.8
        plt.bar(
            x,
            y,
            tick_label=[s.replace(",", "\n") for s in sorted_keys],
            width=width,
        )
        plt.xticks(plt.xticks()[0] + width / 2)
        plt.title(field, fontdict=dict(fontsize=24))


# navg_time=10; plot_timings=True; include_GPU=True; include_CPU=True; print_timings=True; phasings=['real', 'complex']


def bench_mri_2d_nocoils_nofieldmap(
    navg_time=(4, 16),
    plot_timings=False,
    include_GPU=False,
    include_CPU=True,
    print_timings=True,
    phasings=["real", "complex"],
    save_dir=None,
):

    recon_cases = []
    if include_GPU:
        recon_cases += ["GPU,Sp", "GPU,Tab"]
        precisions = ["single"]
    else:
        precisions = ["single", "double"]
    if include_CPU:
        recon_cases += ["CPU,Tab", "CPU,Sp"]

    t = _test_mri_multi(
        ndim=2,
        N0=64,
        grid_os_factor=1.5,
        J0=4,
        Ld=4096,
        n_coils=1,
        fieldmap_segments=None,
        precisions=precisions,
        phasings=phasings,
        recon_cases=recon_cases,
        compare_to_exact=False,
        navg_time=navg_time,
    )

    t_create, t_forward, t_adj, t_norm = _split_timings(t, print_timings)
    if plot_timings:
        axes = _plot_all_timings(
            t_create,
            t_forward,
            t_adj,
            t_norm,
            title_substring="2D (no coils, nofieldmap)",
        )
    else:
        axes = None

    if save_dir is not None:
        fname_base = "bench_2d_001coil_001fmapseg"
        _save_bench(
            save_dir, fname_base, [t_create, t_forward, t_adj, t_norm], axes
        )


def bench_mri_2d_16coils_nofieldmap(
    navg_time=(4, 16),
    plot_timings=False,
    include_GPU=False,
    include_CPU=True,
    print_timings=True,
    phasings=["real", "complex"],
    save_dir=None,
):

    recon_cases = []
    if include_GPU:
        recon_cases += ["GPU,Sp", "GPU,Tab"]
        precisions = ["single"]
    else:
        precisions = ["single", "double"]
    if include_CPU:
        recon_cases += ["CPU,Tab", "CPU,Sp"]

    t = _test_mri_multi(
        ndim=2,
        N0=128,
        grid_os_factor=1.5,
        J0=4,
        Ld=4096,
        n_coils=16,
        fieldmap_segments=None,
        precisions=precisions,
        phasings=phasings,
        recon_cases=recon_cases,
        compare_to_exact=False,
        navg_time=navg_time,
    )

    t_create, t_forward, t_adj, t_norm = _split_timings(t, print_timings)
    if plot_timings:
        axes = _plot_all_timings(
            t_create,
            t_forward,
            t_adj,
            t_norm,
            title_substring="2D (16 coils, nofieldmap)",
        )
    else:
        axes = None

    if save_dir is not None:
        fname_base = "bench_2d_016coil_001fmapseg"
        _save_bench(
            save_dir, fname_base, [t_create, t_forward, t_adj, t_norm], axes
        )


def bench_mri_2d_16coils_fieldmap(
    navg_time=(2, 8),
    plot_timings=False,
    include_GPU=False,
    include_CPU=True,
    print_timings=True,
    phasings=["real", "complex"],
    save_dir=None,
):

    recon_cases = []
    if include_GPU:
        recon_cases += ["GPU,Sp", "GPU,Tab"]
        precisions = ["single"]
    else:
        precisions = ["single", "double"]
    if include_CPU:
        recon_cases += ["CPU,Tab", "CPU,Sp"]

    t = _test_mri_multi(
        ndim=2,
        N0=64,
        grid_os_factor=1.5,
        J0=4,
        Ld=4096,
        n_coils=16,
        fieldmap_segments=6,
        precisions=precisions,
        phasings=phasings,
        recon_cases=recon_cases,
        compare_to_exact=False,
        navg_time=4,
    )

    t_create, t_forward, t_adj, t_norm = _split_timings(t, print_timings)
    if plot_timings:
        axes = _plot_all_timings(
            t_create,
            t_forward,
            t_adj,
            t_norm,
            title_substring="2D (16 coils, 6 fieldmap seg)",
        )
    else:
        axes = None

    if save_dir is not None:
        fname_base = "bench_2d_016coil_006fmapseg"
        _save_bench(
            save_dir, fname_base, [t_create, t_forward, t_adj, t_norm], axes
        )


def bench_mri_2d_16coils_fieldmap_multispectral(
    navg_time=(1, 4),
    plot_timings=False,
    include_GPU=False,
    include_CPU=True,
    print_timings=True,
    phasings=["real", "complex"],
    save_dir=None,
    nspectra=2,
):

    recon_cases = []
    if include_GPU:
        recon_cases += ["GPU,Sp", "GPU,Tab"]
        precisions = ["single"]
    else:
        precisions = ["single", "double"]
    if include_CPU:
        recon_cases += ["CPU,Tab", "CPU,Sp"]
    # spectral_offsets = [0, 434.316]
    spectral_offsets = np.asarray(
        [0, 434.316, 332.124, 485.412, -76.644, 49.8186, 247.8156]
    )
    spectral_offsets = spectral_offsets[:nspectra]

    t = _test_mri_multi(
        ndim=2,
        N0=64,
        grid_os_factor=1.5,
        J0=4,
        Ld=4096,
        n_coils=16,
        fieldmap_segments=6,
        precisions=precisions,
        phasings=phasings,
        recon_cases=recon_cases,
        compare_to_exact=False,
        navg_time=4,
        spectral_offsets=spectral_offsets,
    )

    t_create, t_forward, t_adj, t_norm = _split_timings(t, print_timings)
    if plot_timings:
        title_string = "2D (16 coils, 6 fieldmap seg, %d spectral)" % nspectra
        axes = _plot_all_timings(
            t_create, t_forward, t_adj, t_norm, title_substring=title_string
        )
    else:
        axes = None

    if save_dir is not None:
        fname_base = "bench_2d_016coil_006fmapseg_%dspectra" % nspectra
        _save_bench(
            save_dir, fname_base, [t_create, t_forward, t_adj, t_norm], axes
        )


def bench_mri_3d_nocoils_nofieldmap(
    navg_time=(2, 8),
    plot_timings=False,
    include_GPU=False,
    include_CPU=True,
    print_timings=True,
    phasings=["real", "complex"],
    save_dir=None,
):

    recon_cases = []
    if include_GPU:
        recon_cases += ["GPU,Sp", "GPU,Tab"]
        precisions = ["single"]
    else:
        precisions = ["single", "double"]
    if include_CPU:
        recon_cases += ["CPU,Tab", "CPU,Sp"]

    t = _test_mri_multi(
        ndim=3,
        N0=64,
        grid_os_factor=1.5,
        J0=4,
        Ld=4096,
        n_coils=1,
        fieldmap_segments=None,
        precisions=precisions,
        phasings=phasings,
        recon_cases=recon_cases,
        compare_to_exact=False,
        navg_time=navg_time,
    )

    t_create, t_forward, t_adj, t_norm = _split_timings(t, print_timings)
    if plot_timings:
        axes = _plot_all_timings(
            t_create,
            t_forward,
            t_adj,
            t_norm,
            title_substring="3D (no coils, nofieldmap)",
        )
    else:
        axes = None

    if save_dir is not None:
        fname_base = "bench_3d_001coil_001fmapseg"
        _save_bench(
            save_dir, fname_base, [t_create, t_forward, t_adj, t_norm], axes
        )


def bench_mri_3d_nocoils_fieldmap(
    navg_time=(2, 8),
    plot_timings=False,
    include_GPU=False,
    include_CPU=True,
    print_timings=True,
    phasings=["real", "complex"],
    save_dir=None,
):

    recon_cases = []
    if include_GPU:
        recon_cases += ["GPU,Sp", "GPU,Tab"]
        precisions = ["single"]
    else:
        precisions = ["single", "double"]
    if include_CPU:
        recon_cases += ["CPU,Tab", "CPU,Sp"]

    t = _test_mri_multi(
        ndim=3,
        N0=64,
        grid_os_factor=1.5,
        J0=4,
        Ld=4096,
        n_coils=1,
        fieldmap_segments=6,
        precisions=precisions,
        phasings=phasings,
        recon_cases=recon_cases,
        compare_to_exact=False,
        navg_time=navg_time,
    )

    t_create, t_forward, t_adj, t_norm = _split_timings(t, print_timings)
    if plot_timings:
        axes = _plot_all_timings(
            t_create,
            t_forward,
            t_adj,
            t_norm,
            title_substring="3D (no coils, 6 fieldmap seg)",
        )
    else:
        axes = None

    if save_dir is not None:
        fname_base = "bench_3d_001coil_006fmapseg"
        _save_bench(
            save_dir, fname_base, [t_create, t_forward, t_adj, t_norm], axes
        )


def bench_mri_3d_16coils_nofieldmap(
    navg_time=(1, 4),
    plot_timings=False,
    include_GPU=False,
    include_CPU=True,
    print_timings=True,
    phasings=["real", "complex"],
    save_dir=None,
):

    recon_cases = []
    if include_GPU:
        recon_cases += ["GPU,Sp", "GPU,Tab"]
        precisions = ["single"]
    else:
        precisions = ["single", "double"]
    if include_CPU:
        recon_cases += ["CPU,Tab", "CPU,Sp"]

    t = _test_mri_multi(
        ndim=3,
        N0=64,
        grid_os_factor=1.5,
        J0=4,
        Ld=4096,
        n_coils=16,
        fieldmap_segments=None,
        precisions=precisions,
        phasings=phasings,
        recon_cases=recon_cases,
        compare_to_exact=False,
        navg_time=navg_time,
    )

    t_create, t_forward, t_adj, t_norm = _split_timings(t, print_timings)
    if plot_timings:
        axes = _plot_all_timings(
            t_create,
            t_forward,
            t_adj,
            t_norm,
            title_substring="3D (no coils, nofieldmap)",
        )
    else:
        axes = None

    if save_dir is not None:
        fname_base = "bench_3d_016coil_001fmapseg"
        _save_bench(
            save_dir, fname_base, [t_create, t_forward, t_adj, t_norm], axes
        )


def bench_mri_3d_16coils_fieldmap(
    navg_time=(1, 4),
    plot_timings=False,
    include_GPU=False,
    include_CPU=True,
    print_timings=True,
    phasings=["real", "complex"],
    save_dir=None,
):

    recon_cases = []
    if include_GPU:
        recon_cases += ["GPU,Sp", "GPU,Tab"]
        precisions = ["single"]
    else:
        precisions = ["single", "double"]
    if include_CPU:
        recon_cases += ["CPU,Tab", "CPU,Sp"]

    t = _test_mri_multi(
        ndim=3,
        N0=64,
        grid_os_factor=1.5,
        J0=4,
        Ld=4096,
        n_coils=16,
        fieldmap_segments=6,
        precisions=precisions,
        phasings=phasings,
        recon_cases=recon_cases,
        compare_to_exact=False,
        navg_time=navg_time,
    )

    t_create, t_forward, t_adj, t_norm = _split_timings(t, print_timings)
    if plot_timings:
        axes = _plot_all_timings(
            t_create,
            t_forward,
            t_adj,
            t_norm,
            title_substring="3D (16 coils, 6 fieldmap seg)",
        )
    else:
        axes = None

    if save_dir is not None:
        fname_base = "bench_3d_016coil_006fmapseg"
        _save_bench(
            save_dir, fname_base, [t_create, t_forward, t_adj, t_norm], axes
        )


# plot_timings=True; include_GPU=True; include_CPU=True; print_timings=True; phasings=['real', 'complex']; save_dir='/tmp'


def run_all_bench(
    plot_timings=True,
    include_GPU=True,
    include_CPU=True,
    print_timings=True,
    phasings=["real", "complex"],
    save_dir="/tmp",
):
    common_kwargs = dict(
        plot_timings=plot_timings,
        include_GPU=include_GPU,
        include_CPU=include_CPU,
        print_timings=print_timings,
        phasings=phasings,
        save_dir=save_dir,
    )
    if include_GPU:
        import cupy
        from cupy.cuda import device

        device.get_cusparse_handle()
        cupy.fft.cache.enable()

    # 2d cases
    bench_mri_2d_nocoils_nofieldmap(navg_time=(4, 20), **common_kwargs)
    bench_mri_2d_16coils_nofieldmap(navg_time=(2, 8), **common_kwargs)
    bench_mri_2d_16coils_fieldmap(navg_time=(2, 8), **common_kwargs)
    bench_mri_2d_16coils_fieldmap_multispectral(
        navg_time=4, nspectra=2, **common_kwargs
    )

    # 3d cases
    bench_mri_3d_nocoils_nofieldmap(navg_time=(2, 4), **common_kwargs)
    bench_mri_3d_nocoils_fieldmap(navg_time=(2, 4), **common_kwargs)
    bench_mri_3d_16coils_nofieldmap(navg_time=(1, 4), **common_kwargs)
    bench_mri_3d_16coils_fieldmap(navg_time=(1, 4), **common_kwargs)



"""Coil compression, pre-whitening and sensitivy map estimation utilities.

Much of this code is modified from idmrmrd-python-tools
https://github.com/ismrmrd/ismrmrd-python-tools

Much code that was 2D only has been updated to work with 3D (or nD) data.
The user can specify which axis of the array corresponds to the coils.

The function calculate_csm_inati_iter is an adaptation of a coil sensitivity
map estimation function by Souheil Inati that is included in Gadgetron.

"""

import numpy as np
from scipy import ndimage as ndi
from mrrt.utils._cupy import get_array_module


__all__ = ["apply_csm", "calculate_csm_inati"]


def smooth(img, box=5, xp=None):
    """Smooths coil images

    :param img: Input complex images, ``[y, x] or [z, y, x]``
    :param box: Smoothing block size (default ``5``)

    :returns simg: Smoothed complex image ``[y,x] or [z,y,x]``
    """
    xp, on_gpu = get_array_module(img, xp)
    if xp is np:
        t_real = np.zeros(img.shape, dtype=img.real.dtype)
        t_imag = np.zeros(img.shape, dtype=img.imag.dtype)

        ndi.filters.uniform_filter(img.real, size=box, output=t_real)
        ndi.filters.uniform_filter(img.imag, size=box, output=t_imag)
        simg = t_real + 1j * t_imag
    else:
        try:
            from cupyimg import convolve_separable
        except ImportError:
            # TODO: allow fallback to cupyx.scipy.ndimage when not found
            raise ImportError(
                "GPU-based smoothing requires the cupyimg package."
            )

        if np.isscalar(box):
            box = [box] * img.ndim
        boxfilt = [np.ones(wsize, dtype=img.real.dtype) for wsize in box]
        boxfilt = [w / w.sum() for w in boxfilt]
        boxfilt = [xp.asarray(f) for f in boxfilt]
        simg = convolve_separable(img, boxfilt)
    return simg


def calculate_csm_global(data, roi_mask=None, xp=None):

    xp, on_gpu = get_array_module(data, xp)
    if not ((data.ndim == 3) or (data.ndim == 4)):
        raise ValueError("Data dimension error: data must be 3D or 4D")

    nc = data.shape[-1]

    if roi_mask is not None:
        beta = []
        for ch in range(nc):
            beta.append(data[..., ch][roi_mask].sum())
        beta = xp.asarray(beta)
    else:
        beta = xp.sum(data, axis=tuple(range(data.ndim - 1)))

    beta /= xp.linalg.norm(beta)
    cs = beta.reshape((1,) * (data.ndim - 1) + (nc,)) * xp.ones(
        data.shape, data.dtype
    )

    comim = xp.squeeze((xp.conj(cs) * data).sum(axis=-1))

    rho = xp.sqrt(xp.mean(xp.abs(comim) ** 2, axis=tuple(range(data.ndim - 1))))

    return cs, rho, comim


def calculate_csm_inati(
    data, smoothing=5, niter=5, verbose=True, roi_mask=None, xp=None
):
    """ Fast, iterative coil map estimation for 2D or 3D acquisitions.

    Parameters
    ----------
    im : (..., coil) ndarray
        Input images, (x, y, coil) or (x, y, z, coil).
    smoothing : int or array-like, optional
        Smoothing block size(s) for the spatial axes.
    niter : int, optional
        Maximal number of iterations to run.
    thresh : float, optional
        Threshold on the relative coil map change required for early
        termination of iterations.  If ``thresh=0``, the threshold check
        will be skipped and all `niter` iterations will be performed.
    verbose : bool, optional
        If true, progress information will be printed out at each iteration.

    Returns
    -------
    coil_map : (..., coil) array
        Relative coil sensitivity maps, (x, y, coil) or (x, y, z, coil).
    rho : array
        TODO
    coil_combined : array
        The coil combined image volume, (x, y) or (x, y, z).

    Notes
    -----
    The implementation corresponds to the algorithm described in [1]_ and is a
    port of Gadgetron's `coil_map_3d_Inati_Iter` routine.

    For non-isotropic voxels it may be desirable to use non-uniform smoothing
    kernel sizes, so a length 3 array of smoothings is also supported.

    References
    ----------
    .. [1] S Inati, MS Hansen, P Kellman.  A Fast Optimal Method for Coil
        Sensitivity Estimation and Adaptive Coil Combination for Complex
        Images.  In: ISMRM proceedings; Milan, Italy; 2014; p. 4407.
    """
    xp, on_gpu = get_array_module(data, xp)
    data = xp.ascontiguousarray(data)

    if not ((data.ndim == 3) or (data.ndim == 4)):
        raise ValueError("Data dimension error: data must be 3D or 4D")

    if not (xp.all(xp.asarray(smoothing) > 0)):
        raise ValueError("Box size error: box must be a positive integer")

    if xp.isscalar(smoothing):
        smoothing = (smoothing,) * (data.ndim - 1) + (1,)
    else:
        # add singleton on channel dimension
        smoothing = np.concatenate((np.asarray(smoothing), (1,)), axis=0)
        if smoothing.size != data.ndim:
            raise ValueError(
                "smoothing kernel must match the number of " "image dimensions"
            )

    # initialize
    (cs_global, rho, comim) = calculate_csm_global(data, roi_mask=roi_mask)
    comim = comim

    eps = xp.finfo(data.dtype).eps * xp.abs(data).mean()
    for iter in range(niter):
        if verbose:
            print(
                "Inati coilmap estimation: iteration {} of {}".format(
                    iter + 1, niter
                )
            )

        # comim is D*v, (i.e. u*s)
        # rho is s
        # (u^H*s)*D, i.e. s^2 * v^H
        cs = smooth(
            xp.squeeze(xp.conj(comim[..., xp.newaxis]) * data), smoothing
        )

        # combine s*s*v, using the global combiner
        comim_glob = apply_csm(cs, cs_global, coil_axis=-1)
        # and remove the phase
        cs *= xp.exp(-1j * xp.angle(comim_glob))[..., xp.newaxis]

        # normalize s*s*v, i.e. compute s*s
        csnorm = xp.abs(cs)
        csnorm *= csnorm
        csnorm = csnorm.sum(-1, keepdims=True)
        csnorm = xp.sqrt(csnorm)
        csnorm += eps
        cs /= csnorm

        # D*v = u*s
        comim = apply_csm(data, cs, coil_axis=-1)

    # compute s
    rho = xp.squeeze(xp.sqrt(csnorm))

    return cs, rho, comim


def apply_csm(img, csm, coil_axis=-1, xp=None):
    """Apply coil sensitivity maps to combine images

    :param img: Input images, ``[coil, y, x]``, or ``[coil, z, y, x]``
    :param csm: Coil sensitivity maps, ``[coil, y, x]``, or ``[coil, z, y, x]``

    :returns comim: Combined image, ``[y, x]`` or ``[z, y, x]``
    """
    xp, on_gpu = get_array_module(img, xp)
    img = xp.asarray(img)
    csm = xp.asarray(csm)
    if img.shape != csm.shape:
        raise ValueError(
            "Images and coil sensitivities must have matching shape"
        )
    comim = xp.squeeze(xp.sum(xp.conj(csm) * img, axis=coil_axis))

    return comim



import numpy as np
import pytest

from mrrt.mri.coils import calculate_prewhitening, prewhiten

from mrrt.utils import config

all_xp = [np]
if config.have_cupy:
    import cupy

    all_xp += [cupy]


@pytest.mark.parametrize("xp", all_xp)
def test_calculate_prewhitening(xp):
    nsamples = 2048
    ncoils = 16
    rstate = xp.random.RandomState(16)
    noise = rstate.randn(nsamples, ncoils)
    w, r = calculate_prewhitening(noise, coil_axis=-1, return_full=True)
    assert w.shape == (ncoils, ncoils)
    assert r.shape == (ncoils, ncoils)

    # r should be close to the identity matrix in this case
    eye = xp.eye(ncoils)
    norm_ratio = xp.linalg.norm(r - eye) / xp.linalg.norm(eye)
    assert norm_ratio < 0.1

    # for white noise, w should be nearly diagonal
    assert xp.linalg.norm(xp.diag(w)) / xp.linalg.norm(w) > 0.95

    w = calculate_prewhitening(noise, coil_axis=-1, return_full=False)
    assert w.shape == (ncoils, ncoils)

    # can use arbitrary noise shape with user-specified coil axis
    noise2 = noise.reshape(nsamples // 4, 4, ncoils)
    new_coil_axis = 1
    noise2 = noise2.swapaxes(-1, new_coil_axis)
    w2 = calculate_prewhitening(
        noise2, coil_axis=new_coil_axis, return_full=False
    )
    xp.testing.assert_array_almost_equal(w, w2)


@pytest.mark.parametrize("xp", all_xp)
def test_prewhiten(xp):
    nsamples = 2048
    ncoils = 16
    rstate = xp.random.RandomState(16)
    noise_data = rstate.randn(nsamples, ncoils)

    # some fake data
    data = xp.arange(64 * 64 * ncoils).reshape((64, 64, ncoils))
    data = data + 1j * data[::-1, :, :]

    dataw, w, r = prewhiten(data, noise_data, coil_axis=-1, coil_axis_noi=-1)
    assert w.shape == (ncoils, ncoils)
    assert r.shape == (ncoils, ncoils)
    assert dataw.shape == data.shape
    assert dataw.dtype == data.dtype

    # r should be close to the identity matrix in this case
    eye = xp.eye(ncoils)
    norm_ratio = xp.linalg.norm(r - eye) / xp.linalg.norm(eye)
    assert norm_ratio < 0.1

    # for white noise, w should be nearly diagonal
    assert xp.linalg.norm(xp.diag(w)) / xp.linalg.norm(w) > 0.95

    # transposed data with coils as first axis
    data_trans = data.transpose()
    dataw, w, r = prewhiten(
        data_trans, noise_data, coil_axis=0, coil_axis_noi=-1
    )
    assert w.shape == (ncoils, ncoils)
    assert r.shape == (ncoils, ncoils)
    assert dataw.shape == data_trans.shape
    assert dataw.dtype == data.dtype

    # noise also transposed
    dataw2, w2, r2 = prewhiten(
        data_trans, noise_data.T, coil_axis=0, coil_axis_noi=0
    )
    # result should be the same as previously
    xp.testing.assert_array_almost_equal(dataw, dataw2)



"""Noise pre-whitening of parallel MRI data.

This code is adapted from idmrmrd-python-tools
https://github.com/ismrmrd/ismrmrd-python-tools

"""

import numpy as np
from mrrt.utils import get_array_module


__all__ = ["apply_prewhitening", "calculate_prewhitening", "prewhiten"]


def calculate_prewhitening(
    noise, coil_axis=-1, scale_factor=1.0, return_full=False, xp=None
):
    """Calculates the noise prewhitening matrix

    Parameters
    ----------
    noise : ndarray
        Input noise data (array or matrix)
    coil_axis : int
        Must correspond to the axis in ``noise`` that corresponds to coils.
    scale_factor : float
        Applied on the noise covariance matrix. Used to adjust for effective
        noise bandwith and difference in sampling rate between noise
        calibration and actual measurement:
        scale_factor = (T_acq_dwell/T_noise_dwell)*NoiseReceiverBandwidthRatio
    return_full : bool
        If True, also return the noise correlation matrix.

    Returns
    -------
    W : ndarray
        Prewhitening matrix (upper triangular) of shape (coil, coil).
        ``data_pw = numpy.dot(data, w)`` where `data` is an (nsamples, ncoils)
        array gives prewhitened data.

    R : ndarray
        noise correlation matrix, shape = (coil, coil).

    References
    ----------
    .. [1] Pruessman KP, Weiger M, Bornert P and Boesiger P.  Advances in
    Sensitivity Encoding with Arbitrary k-Space Trajectories.
    Magn. Reson. Med. 46:638-651
    """
    xp, on_gpu = get_array_module(noise, xp)
    noise = xp.asarray(noise)
    coil_axis = coil_axis % noise.ndim
    ncoils = noise.shape[coil_axis]
    if coil_axis != noise.ndim - 1:
        # coil axis must come last
        noise = xp.swapaxes(noise, -1, coil_axis)
    noise = noise.reshape((noise.size // ncoils, ncoils), order="F")
    M = float(noise.shape[0])
    R = (1 / (M - 1)) * xp.dot(noise.T, xp.conj(noise))
    W = xp.linalg.inv(xp.linalg.cholesky(R))
    W = W.T * xp.sqrt(2) * xp.sqrt(scale_factor)
    if return_full:
        return W, R
    else:
        return W


def apply_prewhitening(data, W, order="F", coil_axis=-1, xp=None):
    """Apply the noise prewhitening matrix.

    Parameters
    ----------
    noise : ndarray
        The data to prewhiten.
    W : ndarray
        Input noise prewhitening matrix. This can be computed from noise-only
        data via ``calculate_prewhitening``.
    coil_axis : int
        The axis in ``data`` containing coils.

    Returns
    -------
    w_data : ndarray
        Prewhitened data.

    References
    ----------
    .. [1] Pruessman KP, Weiger M, Bornert P and Boesiger P.  Advances in
    Sensitivity Encoding with Arbitrary k-Space Trajectories.
    Magn. Reson. Med. 46:638-651
    (2001).
    """
    xp, on_gpu = get_array_module(data, xp)
    data = xp.asanyarray(data)
    W = xp.asanyarray(W)
    coil_axis = coil_axis % data.ndim
    ncoils = data.shape[coil_axis]
    if coil_axis != data.ndim - 1:
        # coil axis must come last
        data = xp.swapaxes(data, -1, coil_axis)
    s = data.shape
    data = data.reshape((-1, ncoils), order="F")
    data = xp.dot(data, W).reshape(s, order="F")
    if coil_axis != -1:
        # restore coil axis back to original position
        data = xp.swapaxes(data, -1, coil_axis)
    return data


def prewhiten(data, noise_cal_data, coil_axis=-1, coil_axis_noi=None, xp=None):
    """Noise prewhitening of multichannel MRI data.

    Parameters
    ----------
    data : ndarray
        The data to prewhiten.
    noise_cal_data : ndarray
        Noise calibration data.
    coil_axis : int, optional
        The axis in ``data`` corresponding to coils. By default, the last axis
        is assumed.
    coil_axis_noi : int, optional
        The axis in ``noise_cal_data`` corresponding to coils. By default, the
        last axis is assumed.
    xp : {np, cupy}
        The array module to use.

    Returns
    -------
    data_prewhite
        The prewhitened data.

    W : xp.ndarray
        The noise prewhitening matrix.

    R : xp.ndarray
        The noise correlatin matrix corresponding to ``noise_cal_data``.
    """
    xp, on_gpu = get_array_module(data, xp)
    if coil_axis_noi is None:
        coil_axis_noi = np.argmin(noise_cal_data.shape)

    if not xp.iscomplexobj(data):
        raise ValueError("data must have a complex dtype")

    if data.shape[coil_axis] != noise_cal_data.shape[coil_axis_noi]:
        raise ValueError(
            "data and noise calibration data must have the "
            "same number of channels."
        )

    W, R = calculate_prewhitening(
        noise_cal_data, coil_axis=coil_axis_noi, return_full=True
    )
    data_prewhite = apply_prewhitening(data, W, coil_axis=coil_axis)
    return data_prewhite, W, R



import numpy as np
from numpy.testing import assert_raises, assert_array_almost_equal

from mrrt.mri.operators._MRI import PixelBasis


def _kspace_cartesian(shape, fov):
    if len(fov) != len(shape):
        raise ValueError("shape mismatch")

    omegas = [(np.arange(s) / s - 0.5) * 2 * np.pi for s in shape]
    omegas_grid = np.meshgrid(*omegas, indexing="ij")
    omega = np.stack([o.ravel() for o in omegas_grid], axis=-1)

    # convert to physical units
    kspace = omega.copy()
    for d in range(kspace.shape[-1]):
        dx = fov[d] / shape[d]
        kspace[:, d] /= 2 * np.pi * dx

    return kspace, omega


def test_PixelBasis(show_figure=False):
    fov = np.asarray([240, 240])
    Nd = np.asarray([64, 64])
    dx = fov / Nd
    kspace, omega = _kspace_cartesian(shape=Nd, fov=fov)

    # try both 'dirac' and 'rect'
    b = PixelBasis(kspace, "dirac", fov=fov, Nd=Nd)
    b2 = PixelBasis(kspace, "rect", fov=fov, Nd=Nd)
    b3 = PixelBasis(kspace, "sinc", fov=fov, Nd=Nd)
    assert b.transform is None
    assert b2.transform is not None
    assert b3.transform is not None

    if show_figure:
        import matplotlib.pyplot as plt

        plt.figure()
        plt.subplot(131), plt.plot(kspace[:, 0], kspace[:, 1], "k."),
        plt.xlabel("kx"), plt.ylabel("ky"), plt.axis("image")
        plt.subplot(132), plt.imshow(
            b3.transform.reshape(Nd) / b3.transform.max(), vmin=0, vmax=1
        )
        plt.title("sinc")
        plt.subplot(133), plt.imshow(
            b2.transform.reshape(Nd) / b2.transform.max(), vmin=0, vmax=1
        )
        plt.title("rect")

    # uncrecognized basis
    assert_raises(ValueError, PixelBasis, kspace, "rect2", fov=fov, Nd=Nd)

    # test alternate input argument forms
    b3_v2 = PixelBasis(kspace, "sinc", fov=fov, dx=dx)
    b3_v3 = PixelBasis(kspace, "sinc", fov=fov, dx=dx[0])
    assert_array_almost_equal(b3.transform, b3_v2.transform)
    assert_array_almost_equal(b3.transform, b3_v3.transform)



import itertools

import numpy as np
import pytest

from mrrt.mri import mri_exp_mult
from mrrt.utils import config

all_xp = [np]
if config.have_cupy:
    import cupy

    all_xp += [cupy]

__all__ = ["test_mri_exp_mult"]


@pytest.mark.parametrize(
    "xp, dtype", itertools.product(all_xp, [np.float32, np.float64])
)
def test_mri_exp_mult(xp, dtype):
    """
    test_mri_exp_mult
    """
    L = 1000
    N = 3000
    M = 200
    rstate = xp.random.RandomState(0)
    A = rstate.randn(N, L).astype(dtype)
    A = A + 1j * rstate.randn(N, L).astype(dtype)
    ur = rstate.randn(N).astype(dtype)
    ui = rstate.randn(N).astype(dtype)
    vr = rstate.randn(M).astype(dtype)
    vi = rstate.randn(M).astype(dtype)
    u = ur.astype(dtype) + 1j * ui.astype(dtype)
    v = vr.astype(dtype) + 1j * vi.astype(dtype)

    # Test with ur, v
    y = mri_exp_mult(A, ur, v)
    assert y.dtype == np.promote_types(dtype, np.complex64)

    # Repeat with u, vr
    y = mri_exp_mult(A, u, vr)
    assert y.dtype == np.promote_types(dtype, np.complex64)

    return



import itertools

import time
import numpy as np
import scipy.ndimage as ndi
import pytest

from mrrt.utils import ImageGeometry, ellipse_im
from mrrt.mri import mri_exp_approx

__all__ = ["test_mri_exp_approx"]


def _test_mri_exp_approx1(
    segments=4,
    nx=64,
    tmax=25e-3,
    dt=5e-6,
    autocorr=False,
    use_rmap=True,
    atype="hist,time,unif",
    nhist=None,
    ctest=True,
    verbose=False,
    tol=None,
):
    if verbose:
        from matplotlib import pyplot as plt
        from pyvolplot import subplot_stack

    ti = np.arange(0, tmax, dt)

    if True:
        # Generate a synthetic fieldmap
        fmap = np.zeros((64, 64))
        fmap[6:27, 9:20] = 90
        fmap[36:57, 9:20] = 120
        fmap[5:26, 29:60] = 30
        fmap[37:58, 29:60] = 60
        if nx != 64:
            fmap = ndi.zoom(fmap, nx / 64, order=0)
        kernel_size = int(np.round(5 * nx / 64))
        smoothing_kernel = np.ones((kernel_size, kernel_size)) / (
            kernel_size ** 2
        )
        ndi.convolve(fmap, smoothing_kernel, output=fmap)
        fmap = fmap + 10
        if verbose:
            plt.figure()
            plt.imshow(fmap, interpolation="nearest", cmap="gray")
            plt.title("Field Map")

    if use_rmap:
        # generate a T2 relaxation map
        rmap = (
            np.asarray(
                [[0, 0, 18, 23, 0, 20 * 64 / nx], [6, 0, 8, 8, 0, 3 * 64 / nx]]
            )
            * nx
            / 64
        )
        ig = ImageGeometry(shape=(nx, nx), fov=(nx, nx))
        rmap, params = 1 * ellipse_im(ig, rmap, oversample=3)
        if verbose:
            plt.figure()
            plt.imshow(rmap, cmap="gray", interpolation="nearest")
            plt.title("Relax Map"),
    else:
        rmap = 0

    zmap = rmap + (2j * np.pi) * fmap

    if not nhist:
        if not np.any(rmap > 0):
            nhist = [40]
        else:
            nhist = [40, 10]

    # autocorr_arg = ['autocorr', True] # test autocorrelation version

    if True:  # convert to single precision
        ti = np.asarray(ti, dtype="float32")
        zmap = np.asarray(zmap, dtype="complex64")  # single precision complex

    if isinstance(segments, int):
        pass
    elif isinstance(segments, (list, tuple)) and len(segments) == 2:
        pass
    else:
        raise ValueError("Invalid choice for segments")

    kwargs = {"autocorr": autocorr, "ctest": ctest, "verbose": verbose}
    tstart = time.time()
    if tol is None:
        B, C, hk, zk = mri_exp_approx(
            ti, zmap, segments, approx_type=(atype, nhist), **kwargs
        )
    else:
        B, C, hk, zk = mri_exp_approx(
            ti, zmap, [segments, tol], approx_type=(atype, nhist), **kwargs
        )
    print("\tduration=%g s" % (time.time() - tstart))

    if ctest:
        Eh = np.exp(-ti[:, np.newaxis] * zk.ravel()[np.newaxis, :])
    else:
        Eh = np.exp(-ti[:, np.newaxis] * zmap.ravel()[np.newaxis, :])
    Ep = np.dot(B, C)  # matrix product
    err = np.abs(Eh - Ep)
    mse = np.mean(np.power(err, 2), axis=0)
    if ctest:
        wrms = np.sqrt(np.dot(mse, hk) / np.sum(hk))
    else:
        wrms = -1

    if verbose:
        subplot_stack(1000 * ti, B, title="Basis Components", colors=["k", "m"])

    nf = np.floor(nhist[0] / 4)
    if len(nhist) == 2:
        ik = np.array([0, nf, 2 * nf, 3 * nf, nhist[1] - 1]) + 2 * nf * nhist[1]
        ik = ik.tolist()
    elif len(nhist) == 1:
        ik = [0, nf, 2 * nf, 3 * nf, nhist[0] - 1]
    ik = np.asarray(ik, dtype=int)

    mse_mean = mse.mean()
    max_err = err.max()

    if verbose:
        fig = subplot_stack(1000 * ti, Eh[:, ik], colors=["g", "k"])
        fig = subplot_stack(
            1000 * ti,
            Ep[:, ik],
            colors=["b--", "r--"],
            fig=fig,
            title="True and Approx",
        )
        fig = subplot_stack(
            1000 * ti,
            err[:, ik],
            colors=["b--", "r--"],
            title="True and Approx",
        )
        print(
            "\tfor L=%d, wrms=%g, mse = %g, max_err=%g"
            % (B.shape[1], wrms, mse_mean, max_err)
        )
    return wrms, mse_mean, max_err


@pytest.mark.parametrize(
    "use_rmap, alg",
    itertools.product(
        [False, True],
        [
            "hist,svd",
            "hist,time,unif",
            "time,unif",
            "hist,fs,unif",
            "hist,fs,prctile",
            "hist,fs,kmeans",
        ],
    ),
)
@pytest.mark.filterwarnings("ignore:the matrix subclass is not")
def test_mri_exp_approx(use_rmap, alg, verbose=False):
    if alg == ["hist,fs,kmeans"]:
        pytest.importorskip("sklearn")

    tmax = 25e-3  # overall time duration (s)
    wrms, mse_mean, max_err = _test_mri_exp_approx1(
        segments=8,  # number of segments
        nx=64,
        tmax=tmax,
        dt=1e-3,
        autocorr=False,
        use_rmap=use_rmap,
        atype=alg,
        nhist=None,
        ctest=True,
        verbose=verbose,
    )

    if alg == "hist,fs,prctile" or (use_rmap and alg == "hist,fs,kmeans"):
        # may want to just remove these options, as they perform relatively
        # poorly
        assert mse_mean < 0.01
    else:
        assert mse_mean < 1e-5



import numpy as np

__all__ = ["generate_fieldmap"]


def generate_fieldmap(shape, fmap_peak=80):
    """Return a 2d or 3d Gaussian bump for use as a simulated filedmap.

    Parameters
    ----------
    shape : tuple of int
        The shape of the fieldmap.
    fmap_peak : float, optional
        The maximum magnitude of the fieldmap, in Hz.

    Returns
    -------
    z_map : ndarray
        The field map in complex form as expected by the MRI NonCartesian
        operator. The imaginary component corresponds to the off-resonance map
        in radians. (The real component corresponds to relaxation rather than
        off-resonance and will be zero).

    Notes
    -----
    ``z_map = r_map + 1j * f_map`` where ``r_map`` is a T2* relaxation map and
    ``f_map`` is a fieldmap in radians.

    """
    ndim = len(shape)
    if ndim < 2 or ndim > 3:
        raise ValueError("Only 2D and 3D cases supported")
    # make up a fake gaussian fieldmap...
    a = 2.5
    k1 = np.arange(-(shape[0] - 1) / 2, (shape[0] - 1) / 2 + 1)
    w1 = np.exp(-0.5 * (a * k1 / (shape[0] / 2)) ** 2)
    k2 = np.arange(-(shape[1] - 1) / 2, (shape[1] - 1) / 2 + 1)
    w2 = np.exp(-0.5 * (a * k2 / (shape[1] / 2)) ** 2)

    if ndim == 3:
        k3 = np.arange(-(shape[2] - 1) / 2, (shape[2] - 1) / 2 + 1)
        w3 = np.exp(-0.5 * (a * k3 / (shape[2] / 2)) ** 2)
        w1 = w1[:, None, None]
        w2 = w2[None, :, None]
        w3 = w3[None, None, :]
        z_map = 0 + 1j * 2 * np.pi * fmap_peak * (w1 * w2 * w3)
    elif ndim == 2:
        w1 = w1[:, None]
        w2 = w2[None, :]
        z_map = 0 + 1j * 2 * np.pi * fmap_peak * (w1 * w2)

    return z_map



"""MRI test objects created from basic shapes with analytical expression.

These are used in testing and benchmarking scripts as well as some demos.

The following shapes are implemented based on analytical expressions of their
Fourier transform:

    Gaussians
    Rectangles
    Cylinders
    Diracs

The class AnalyticalComposite is used to generate a composite phantom built up
of analytical objects of the above type.

The following functions generate specific test phantoms as used in various
testing and benchmarking scripts:

    mri_object_1d
    mri_object_2d
    mri_object_2d_multispectral
    mri_object_3d

Much of the code in this module is ported from equivalent Matlab code by
Jeff Fessler and his students (Michigan Image Reconstruction Toolbox).
"""
from abc import ABC, abstractmethod
import warnings

import numpy as np

from mrrt.utils import rect, jinc


__all__ = [
    "AnalyticalComposite",
    "Cylinders",
    "Diracs",
    "Gaussians",
    "Rectangles",
    "mri_object_1d",
    "mri_object_2d",
    "mri_object_2d_multispectral",
    "mri_object_3d",
]


class Fourier_Analytical(ABC):

    fixup = False

    def __init__(self, ndim):
        self.ndim = ndim

    def _fixup_shape(self, x, d, fixup=None):
        if fixup is None:
            fixup = self.fixup
        if fixup and x.ndim != self.ndim:
            if x.ndim == 1:
                new_shape = [1] * self.ndim
                new_shape[d] = x.size
                x = x.reshape(new_shape)
            else:
                raise ValueError(
                    "Each coordinate array must be 1d or have full" "shape."
                )
        return x

    @abstractmethod
    def image(self, *coords):
        """Image domain representation."""

    @abstractmethod
    def kspace(self, *coords):
        """Fourier domain representation not implemented."""


class Gaussians(Fourier_Analytical):
    def __init__(
        self,
        amplitudes,
        centers,
        widths,
        dtype=np.float64,
        fixup=False,
        offsets=None,
    ):
        centers = np.asarray(centers)
        amplitudes = np.asarray(amplitudes)
        widths = np.asarray(widths)
        if amplitudes.ndim != 1:
            raise ValueError("amplitudes should be a shape (N, ) array")
        if centers.ndim != 2:
            raise ValueError("centers should be a shape (N, ndim) array")
        if widths.ndim != 2:
            raise ValueError("centers should be a shape (N, ndim) array")
        if (centers.shape[0] != amplitudes.shape[0]) or (
            widths.shape[0] != amplitudes.shape[0]
        ):
            raise ValueError("shape mismatch")
        self.n = amplitudes.size
        self.ndim = centers.shape[1]
        self.amplitudes = amplitudes
        self.centers = centers
        self.widths = widths
        self.dtype = dtype
        self.fixup = fixup
        self.offsets = offsets
        if self.offsets is None:
            # set of tuples for each component where the first element of the
            # tumple is the relative amplitude of the component and the second
            # is it's frequency offset.
            self.offsets = [0] * self.n
        self.unique_offsets = np.unique(self.offsets)
        self.n_offsets = self.unique_offsets.size

    def image(self, *coords, squeeze=True):
        # Evaluate values in the image domain.
        if len(coords) != self.ndim:
            raise ValueError(
                "The number of coordinate arrays must equal the number of "
                "dimensions"
            )
        denom = np.sqrt(np.log(256)) / np.sqrt(2 * np.pi)
        widths = self.widths / denom
        if self.n_offsets == 1 and squeeze:
            out = 0.0
            for n in range(self.n):
                g = 1.0
                for d in range(self.ndim):
                    x = self._fixup_shape(coords[d], d)
                    c = self.centers[n, d]
                    w = widths[n, d]
                    g = g * np.exp(-np.pi * ((x - c) / w) ** 2)
                out += self.amplitudes[n] * g
        else:
            all_out = [0] * self.n_offsets
            for m, off in enumerate(self.unique_offsets):
                out = 0
                for n in np.where(self.offsets == off)[0]:
                    g = 1.0
                    for d in range(self.ndim):
                        x = self._fixup_shape(coords[d], d)
                        c = self.centers[n, d]
                        w = widths[n, d]
                        g = g * np.exp(-np.pi * ((x - c) / w) ** 2)
                    out += self.amplitudes[n] * g
                all_out[m] = out
            out = np.stack(all_out, axis=-1)

        return np.asarray(out, dtype=self.dtype)

    def kspace(self, *coords, t=None):
        # Evaluate values in the Fourier domain.
        if len(coords) != self.ndim:
            raise ValueError(
                "The number of coordinate arrays must equal the number of "
                "dimensions"
            )
        denom = np.sqrt(np.log(256)) / np.sqrt(2 * np.pi)
        widths = self.widths / denom
        out = 0.0
        for n in range(self.n):
            g = 1.0
            omega = 2 * np.pi * self.offsets[n]
            for d in range(self.ndim):
                x = self._fixup_shape(coords[d], d)
                c = self.centers[n, d]
                w = widths[n, d]
                g = g * w
                g = g * np.exp(-np.pi * (x * w) ** 2)
                g = g * np.exp(-2j * np.pi * (x * c))
            if omega == 0:
                out += self.amplitudes[n] * g
            else:
                if t is None:
                    raise ValueError(
                        "for non-zero offsets, a time vector "
                        "corresponding to the trajectory must be "
                        "provided"
                    )
                out += self.amplitudes[n] * g * np.exp(1j * omega * t)
        return np.asarray(out, dtype=np.result_type(self.dtype, np.complex64))


class Rectangles(Fourier_Analytical):
    def __init__(
        self, amplitudes, centers, widths, dtype=np.float64, offsets=None
    ):
        centers = np.asarray(centers)
        amplitudes = np.asarray(amplitudes)
        widths = np.asarray(widths)
        if amplitudes.ndim != 1:
            raise ValueError("amplitudes should be a shape (N, ) array")
        if centers.ndim != 2:
            raise ValueError("centers should be a shape (N, ndim) array")
        if widths.ndim != 2:
            raise ValueError("centers should be a shape (N, ndim) array")
        if (centers.shape[0] != amplitudes.shape[0]) or (
            widths.shape[0] != amplitudes.shape[0]
        ):
            raise ValueError("shape mismatch")
        self.n = amplitudes.size
        self.ndim = centers.shape[1]
        self.amplitudes = amplitudes
        self.centers = centers
        self.widths = widths
        self.dtype = dtype
        self.offsets = offsets
        if self.offsets is None:
            # set of tuples for each component where the first element of the
            # tumple is the relative amplitude of the component and the second
            # is it's frequency offset.
            self.offsets = [0] * self.n
        self.unique_offsets = np.unique(self.offsets)
        self.n_offsets = self.unique_offsets.size

    def image(self, *coords, squeeze=True):
        # Evaluate values in the image domain.
        if len(coords) != self.ndim:
            raise ValueError(
                "The number of coordinate arrays must equal the number of "
                "dimensions"
            )
        if self.n_offsets == 1 and squeeze:
            out = 0.0
            for n in range(self.n):
                g = 1.0
                for d in range(self.ndim):
                    x = self._fixup_shape(coords[d], d)
                    c = self.centers[n, d]
                    w = self.widths[n, d]
                    g = g * rect((x - c) / w)
                out += self.amplitudes[n] * g
        else:
            out = [0] * self.n_offsets
            for m, off in enumerate(self.unique_offsets):
                o = 0
                for n in np.where(self.offsets == off)[0]:
                    g = 1.0
                    for d in range(self.ndim):
                        x = self._fixup_shape(coords[d], d)
                        c = self.centers[n, d]
                        w = self.widths[n, d]
                        g = g * rect((x - c) / w)
                    o += self.amplitudes[n] * g
                out[m] = o
            out = np.stack(out, axis=-1)
        return np.asarray(out, dtype=self.dtype)

    def kspace(self, *coords, t=None):
        # Evaluate values in the Fourier domain.
        if len(coords) != self.ndim:
            raise ValueError(
                "The number of coordinate arrays must equal the number of "
                "dimensions"
            )
        out = 0.0
        for n in range(self.n):
            omega = 2 * np.pi * self.offsets[n]
            g = 1.0
            ph = 0.0
            for d in range(self.ndim):
                f = self._fixup_shape(coords[d], d)
                c = self.centers[n, d]
                w = self.widths[n, d]
                sinc_term = w * np.sinc(f * w)
                g = g * sinc_term
                ph = ph + f * c
            g = g * np.exp(-2j * np.pi * ph)
            if omega == 0:
                out += self.amplitudes[n] * g
            else:
                if t is None:
                    raise ValueError(
                        "for non-zero offsets, a time vector "
                        "corresponding to the trajectory must be "
                        "provided"
                    )
                out += self.amplitudes[n] * g * np.exp(1j * omega * t)
        return np.asarray(out, dtype=np.result_type(self.dtype, np.complex64))


class Cylinders(Fourier_Analytical):
    def __init__(
        self,
        amplitudes,
        centers,
        radii,
        heights=None,
        dtype=np.float64,
        offsets=None,
    ):
        centers = np.asarray(centers)
        amplitudes = np.asarray(amplitudes)
        radii = np.asarray(radii)
        if amplitudes.ndim != 1:
            raise ValueError("amplitudes should be a shape (N, ) array")
        if radii.ndim != 1:
            raise ValueError("radii should be a shape (N, ) array")
        if centers.ndim != 2:
            raise ValueError("centers should be a shape (N, ndim) array")
        if (centers.shape[0] != amplitudes.shape[0]) or (
            radii.shape[0] != amplitudes.shape[0]
        ):
            raise ValueError("shape mismatch")
        if heights is not None:
            heights = np.asarray(heights)
            if heights.ndim != 1:
                raise ValueError("heights should be a shape (N, ) array")
            if heights.shape[0] != amplitudes.shape[0]:
                raise ValueError("shape mismatch")
            if centers.shape[1] != 3:
                raise ValueError("heights requires 3d centers")
        self.n = amplitudes.size
        self.ndim = centers.shape[1]
        self.amplitudes = amplitudes
        self.centers = centers
        self.radii = radii
        self.heights = heights
        self.dtype = dtype
        self.offsets = offsets
        if self.offsets is None:
            # set of tuples for each component where the first element of the
            # tumple is the relative amplitude of the component and the second
            # is it's frequency offset.
            self.offsets = [0] * self.n
        self.unique_offsets = np.unique(self.offsets)
        self.n_offsets = self.unique_offsets.size

    def image(self, *coords, squeeze=True):
        # Evaluate values in the image domain.
        if len(coords) != self.ndim:
            raise ValueError(
                "The number of coordinate arrays must equal the number of "
                "dimensions"
            )

        if self.heights is not None:
            ndim_circ = self.ndim - 1
        else:
            ndim_circ = self.ndim

        if self.n_offsets == 1 and squeeze:
            out = 0.0
            for n in range(self.n):
                rr = 0.0
                for d in range(ndim_circ):
                    x = self._fixup_shape(coords[d], d)
                    c = self.centers[n, d]
                    rr = rr + (x - c) ** 2
                circ = rr < self.radii[n] ** 2
                if self.heights is None:
                    out += self.amplitudes[n] * circ
                else:
                    z = self._fixup_shape(coords[-1], self.ndim - 1)
                    zc = self.centers[n, -1]
                    zh = self.heights[n]
                    out += self.amplitudes[n] * circ * rect((z - zc) / zh)
        else:
            out = [0] * self.n_offsets
            for m, off in enumerate(self.unique_offsets):
                o = 0
                for n in np.where(self.offsets == off)[0]:
                    rr = 0.0
                    for d in range(ndim_circ):
                        x = self._fixup_shape(coords[d], d)
                        c = self.centers[n, d]
                        rr = rr + (x - c) ** 2
                    circ = rr < self.radii[n] ** 2
                    if self.heights is None:
                        out += self.amplitudes[n] * circ
                    else:
                        z = self._fixup_shape(coords[-1], self.ndim - 1)
                        zc = self.centers[n, -1]
                        zh = self.heights[n]
                        o += self.amplitudes[n] * circ * rect((z - zc) / zh)
                out[m] = o
            out = np.stack(out, axis=-1)
        return np.asarray(out, dtype=self.dtype)

    def kspace(self, *coords, t=None):
        # Evaluate values in the Fourier domain.
        # Evaluate values in the image domain.
        if len(coords) != self.ndim:
            raise ValueError(
                "The number of coordinate arrays must equal the number of "
                "dimensions"
            )
        if self.ndim == 2:
            # disc
            u, v = coords[0], coords[1]
        if self.ndim == 3:
            # cylinder
            u, v, w = coords[0], coords[1], coords[2]
            w = self._fixup_shape(w, 2)
        u = self._fixup_shape(u, 0)
        v = self._fixup_shape(v, 1)

        out = 0.0
        for n in range(self.n):
            omega = 2 * np.pi * self.offsets[n]
            r = self.radii[n]
            jinc_term = r ** 2 * 4 * jinc(2 * np.sqrt(u ** 2 + v ** 2) * r)
            if self.heights is not None:
                zh = self.heights[n]
                sinc_term = zh * np.sinc(w * zh)
                xc, yc, zc = self.centers[n]
                ph = xc * u + yc * v + zc * w
                o = (
                    self.amplitudes[n]
                    * jinc_term
                    * sinc_term
                    * np.exp(-2j * np.pi * ph)
                )
            else:
                xc, yc = self.centers[n]
                ph = xc * u + yc * v
                o = self.amplitudes[n] * jinc_term * np.exp(-2j * np.pi * ph)
            if omega == 0:
                out = out + o
            else:
                if t is None:
                    raise ValueError(
                        "for non-zero offsets, a time vector "
                        "corresponding to the trajectory must be "
                        "provided"
                    )
                out = out + o * np.exp(1j * omega * t)
        return np.asarray(out, dtype=np.result_type(self.dtype, np.complex64))


class Diracs(Fourier_Analytical):
    def __init__(self, amplitudes, centers, dtype=np.float64):
        centers = np.asarray(centers)
        amplitudes = np.asarray(amplitudes)
        if amplitudes.ndim != 1:
            raise ValueError("amplitudes should be a shape (N, ) array")
        if centers.ndim != 2:
            raise ValueError("centers should be a shape (N, ndim) array")
        if centers.shape[0] != amplitudes.shape[0]:
            raise ValueError("shape mismatch")
        self.n = amplitudes.size
        self.ndim = centers.shape[1]
        self.amplitudes = amplitudes
        self.centers = centers
        self.dtype = dtype

    def image(self, *coords):
        # Evaluate values in the image domain.
        if len(coords) != self.ndim:
            raise ValueError(
                "The number of coordinate arrays must equal the number of "
                "dimensions"
            )
        out = 0.0
        for n in range(self.n):
            g = 1.0
            for d in range(self.ndim):
                x = self._fixup_shape(coords[d], d)
                c = self.centers[n, d]
                g = g * (x == c)
            out += self.amplitudes[n] * g
        out[out != 0] = np.inf
        warnings.warn("Image of dirac is invalid")
        return np.asarray(out, dtype=self.dtype)

    def kspace(self, *coords):
        # Evaluate values in the Fourier domain.
        if len(coords) != self.ndim:
            raise ValueError(
                "The number of coordinate arrays must equal the number of "
                "dimensions"
            )
        out = 0.0
        for n in range(self.n):
            g = 0.0
            for d in range(self.ndim):
                x = self._fixup_shape(coords[d], d)
                c = self.centers[n, d]
                g = g + x * c
            out += self.amplitudes[n] * np.exp(-2j * np.pi * g)
        return np.asarray(out, dtype=np.result_type(self.dtype, np.complex64))


class AnalyticalComposite(object):
    """
    Object that describes image-domain and Fourier-domain spectra of simple
    structures such as rectangles, disks and Gaussian bumps.

    These functions are useful for simple "idealized" MRI simulations
    where the data is modeled as analytical Fourier samples,
    i.e., no field inhomogeneity and no relaxation effects.
    """

    def __init__(self, objects):

        if isinstance(objects, Fourier_Analytical):
            objects = [objects]

        for obj in objects:
            if not isinstance(obj, Fourier_Analytical):
                raise ValueError(
                    "objects should be a list of analytical Fourier objects"
                )
        self.objects = objects

        # bookkeeping stuff for cases where different object sets may have
        # different numbers of spectral offsets
        tmp = np.concatenate([o.unique_offsets for o in self.objects])
        self.unique_offsets = np.unique(tmp)
        # reorder so offset=0 is always the first component
        self.unique_offsets = np.concatenate(
            (
                self.unique_offsets[self.unique_offsets == 0],
                sorted(self.unique_offsets[self.unique_offsets != 0]),
            )
        )
        self.n_offsets_total = self.unique_offsets.size
        self.img_idx = [
            [
                np.where(self.unique_offsets == o)[0][0]
                for o in obj.unique_offsets
            ]
            for obj in self.objects
        ]

    def image(self, *coords):
        for n, object_set in enumerate(self.objects):
            o = object_set.image(*coords, squeeze=False)
            if n == 0:
                out = np.zeros(
                    o.shape[:-1] + (self.n_offsets_total,), dtype=o.dtype
                )
            out[..., self.img_idx[n]] = out[..., self.img_idx[n]] + o.real
        return np.squeeze(out)

    def kspace(self, *coords, t=None):
        out = 0
        for object_set in self.objects:
            out = out + object_set.kspace(*coords, t=t)
        return out


def mri_object_1d(fov, units="mm", dtype=np.float64):
    if len(fov) > 1:
        fov = fov[0]

    if units == "cm":
        ref_fov = 25.6
    else:
        ref_fov = 256.0  # mm

    rectangles = Rectangles(
        amplitudes=[1, 1, 1],
        centers=((0,), (-50,), (0,)),
        widths=((200,), (40,), (50,)),
        dtype=dtype,
    )
    rectangles.centers = rectangles.centers / ref_fov * np.asarray(fov)
    rectangles.widths = rectangles.widths / ref_fov * np.asarray(fov)

    return AnalyticalComposite([rectangles])


def mri_object_2d(fov, units="mm", dtype=np.float64):

    if units == "cm":
        ref_fov = 25.6
    else:
        ref_fov = 256.0  # mm

    rectangles = Rectangles(
        amplitudes=[1, 1, 1, 1],
        centers=((0, 0), (-50, -50), (50, -50), (0, 50)),
        widths=((200, 200), (40, 40), (20, 20), (50, 50)),
        dtype=dtype,
    )
    rectangles.centers = rectangles.centers / ref_fov * np.asarray(fov)
    rectangles.widths = rectangles.widths / ref_fov * np.asarray(fov)

    gaussians = Gaussians(
        amplitudes=[1, 1, 1, 1, 1, 1, 1, 1],
        centers=(
            (-70, 0),
            (-60, 0),
            (-50, 0),
            (-40, 0),
            (-20, 0),
            (0, 0),
            (20, 0),
            (50, 0),
        ),
        widths=((1, 1), (2, 2), (3, 3), (4, 4), (5, 5), (6, 6), (7, 7), (8, 8)),
        dtype=dtype,
    )
    gaussians.centers = gaussians.centers / ref_fov * np.asarray(fov)
    gaussians.widths = gaussians.widths / ref_fov * np.asarray(fov)

    return AnalyticalComposite([rectangles, gaussians])


def mri_object_2d_multispectral(
    fov, spectral_offsets=None, units="mm", dtype=np.float64
):
    """Returns a pair of analytical objects.

    The first object returned is the object magnitude and the second is the
    spectral offsets.

    Examples
    --------
    from mrrt.utils import ImageGeometry
    import matplotlib.pyplot as plt
    fov = 256
    ig = ImageGeometry(shape=(128, 128), fov=fov)
    o1, o2 = mri_object_2d_multispectral(fov, n_fat_peaks=6)
    fig, axes = plt.subplots(1, 2)
    axes[0].imshow(o1.image(*ig.grid()))
    axes[1].imshow(o2.image(*ig.grid()))
    """

    if units == "cm":
        ref_fov = 25.6
    else:
        ref_fov = 256.0  # mm

    # Generate objects for the magnitude component
    if len(spectral_offsets) > 7:
        raise ValueError("maximum of 7 spectral offsets supported")
    offsets = np.zeros(7)
    offsets[-len(spectral_offsets) :] = spectral_offsets
    rectangles = Rectangles(
        amplitudes=[1, 1, 1, 1, 1, 1, 1],
        offsets=offsets,
        centers=(
            (0, 0),
            (-50, -50),
            (50, -50),
            (0, -50),
            (-50, 50),
            (50, 50),
            (0, 50),
        ),
        widths=(
            (200, 200),
            (30, 30),
            (30, 30),
            (30, 30),
            (30, 30),
            (30, 30),
            (30, 30),
        ),
        dtype=dtype,
    )
    rectangles.centers = rectangles.centers / ref_fov * np.asarray(fov)
    rectangles.widths = rectangles.widths / ref_fov * np.asarray(fov)

    gaussians = Gaussians(
        amplitudes=[1, 1, 1, 1, 1, 1, 1, 1],
        centers=(
            (-70, 0),
            (-60, 0),
            (-50, 0),
            (-40, 0),
            (-20, 0),
            (0, 0),
            (20, 0),
            (50, 0),
        ),
        widths=((1, 1), (2, 2), (3, 3), (4, 4), (5, 5), (6, 6), (7, 7), (8, 8)),
        dtype=dtype,
    )
    gaussians.centers = gaussians.centers / ref_fov * np.asarray(fov)
    gaussians.widths = gaussians.widths / ref_fov * np.asarray(fov)

    return AnalyticalComposite([rectangles, gaussians])


def mri_object_3d(fov, units="mm", dtype=np.float64):

    # cylinders
    cylinders = Cylinders(
        amplitudes=[1],
        centers=[(0, 0, 0)],
        radii=[0.4 * fov[0]],
        heights=[0.9 * fov[2]],
    )

    if units == "cm":
        ref_fov = 25.6
    else:
        ref_fov = 256.0  # mm

    # Rectangles
    rectangles = Rectangles(
        amplitudes=[1, 1, 1],
        centers=((-50, -50, 0), (50, -50, 40), (0, 50, -40)),
        widths=((40, 40, 40), (20, 20, 50), (30, 30, 60)),
        dtype=dtype,
    )
    rectangles.centers = rectangles.centers / ref_fov * np.asarray(fov)
    rectangles.widths = rectangles.widths / ref_fov * np.asarray(fov)

    # Gaussians
    gaussians = Gaussians(
        amplitudes=[1, 1, 1, 1, 1, 1, 1, 1],
        centers=(
            (-70, 0, 0),
            (-60, 0, 0),
            (-50, 0, 0),
            (-40, 0, 0),
            (-20, 0, 0),
            (0, 0, 0),
            (20, 0, 0),
            (50, 0, 0),
        ),
        widths=(
            (1, 1, 1),
            (2, 2, 2),
            (3, 3, 3),
            (4, 4, 4),
            (5, 5, 5),
            (6, 6, 6),
            (7, 7, 7),
            (8, 8, 8),
        ),
        dtype=dtype,
    )
    gaussians.centers = gaussians.centers / ref_fov * np.asarray(fov)
    gaussians.widths = gaussians.widths / ref_fov * np.asarray(fov)
    return AnalyticalComposite([cylinders, rectangles, gaussians])



"""
Multi-peak fat/water models for Magnetic Resonance Imaging.
"""

import numpy as np


__all__ = ["fat_model"]


def six_peak_fat_model(gamma=42.58, field_strength=3, components=6):
    """Six peak fat model.

    Parameters
    ----------
    gamma : float, optional
        The gyromagnetic ratio in MHz/Tesla (default value is for is for 1H).
    field_strengh : float, optional
        MRI field strength (in Tesla).

    Returns
    -------
    relative_frequencies_Hz : array
        A numpy array containing the relative frequencies in Hz for the fat
        peaks.
    fat_rel_amps : array
        The relative amplitude of the fat peaks (the sum of all amplitudes
        is normalized to 1).

    Notes
    -----
    The output arrays are sorted in order of descending relative amplitude.

    These values are taken from the model used by the FWQPBO toolbox and for
    the 2012 ISMRM fat/water workshop challenge.

    The six components in this model are (offsets listed at 3 Tesla):
        Component 1:  Ampltidue 0.693: ,  Offset (Hz): 434.3
        Component 2:  Ampltidue 0.128: ,  Offset (Hz): 332.1
        Component 3:  Ampltidue 0.087: ,  Offset (Hz): 485.4
        Component 4:  Ampltidue 0.048: ,  Offset (Hz): -76.6
        Component 5:  Ampltidue 0.039: ,  Offset (Hz): 49.8
        Component 6:  Ampltidue 0.004: ,  Offset (Hz): 247.8

    These are similar to the six peak models from [1]_, [2]_ and references
    therein.

    References
    ----------
    .. [1] H. Yu, A. Shimakawa, C.A. McKenzie, E. Brodsky, J.H. Brittain,
           S.B. Reeder. Multi-Echo Water-Fat Separation and Simultaneous R2*
           Estimation with Multi-Frequency Fat Spectrum Modeling.
           Magn. Reson. Med. 2008; 60(5):1122-1134.
           DOI:10.1002/mrm.21737

    .. [2] D. Hernando, Z.-P. Liang, P. Kellman.  Chemical Shift-Based
           Water/Fat Separation: A Comparison of Signal Models.
           Magn. Reson. Med. 2010; 64(3):811-822.
           DOI:10.1002/mrm.22455
    """
    water_cs = 4.7  # water chemical shift in ppm

    # fat peak chemical shifts in ppm
    fat_cs = np.asarray([5.3, 4.31, 2.76, 2.1, 1.3, 0.9])

    # relative amplitudes of the peaks
    fat_rel_amps = np.asarray([0.048, 0.039, 0.004, 0.128, 0.693, 0.087])

    relative_shifts = water_cs - fat_cs

    relative_frequencies_Hz = gamma * field_strength * relative_shifts

    # indices to sort in descending order
    si = np.argsort(fat_rel_amps)[::-1]
    fat_rel_amps = fat_rel_amps[si]
    relative_frequencies_Hz = relative_frequencies_Hz[si]

    if components < 1 or components > 6:
        raise ValueError("components must be between 1 and 6")
    if components < 6:
        # truncate and renormalize
        relative_frequencies_Hz = relative_frequencies_Hz[:components]
        fat_rel_amps = fat_rel_amps[:components]

    # make sure sum of amplitudes is 1.0
    fat_rel_amps /= np.sum(fat_rel_amps)

    return -relative_frequencies_Hz, fat_rel_amps


def three_peak_fat_model(gamma=42.58, field_strength=3):
    """Three peak fat model from the liver as given in [1]_.

    Parameters
    ----------
    gamma : float, optional
        The gyromagnetic ratio in MHz/Tesla (default value is for is for 1H).
    field_strengh : float, optional
        MRI field strength (in Tesla).

    Returns
    -------
    relative_frequencies_Hz : array
        A numpy array containing the relative frequencies in Hz for the fat
        peaks.
    fat_rel_amps : array
        The relative amplitude of the fat peaks (the sum of all amplitudes
        is normalized to 1).

    Notes
    -----
    The values used are an average of patient 1 and 2 of Table 1 in [1]_.

    References
    ----------
    .. [1] H. Yu, A. Shimakawa, C.A. McKenzie, E. Brodsky, J.H. Brittain,
           S.B. Reeder. Multi-Echo Water-Fat Separation and Simultaneous R2*
           Estimation with Multi-Frequency Fat Spectrum Modeling.
           Magn. Reson. Med. 2008; 60(5):1122-1134.
           DOI:10.1002/mrm.21737
    """

    # relative amplitudes of the peaks
    fat_rel_amps = np.asarray([0.74, 0.18, 0.08])
    fat_rel_amps /= fat_rel_amps.sum()

    freqs_3T_1H = np.asarray([420, 318, -94])
    relative_frequencies_Hz = gamma / 42.58 * field_strength / 3 * freqs_3T_1H

    return -relative_frequencies_Hz, fat_rel_amps


def single_peak_fat_model(gamma=42.58, field_strength=3):
    """Simple, single peak fat model.

    Parameters
    ----------
    gamma : float, optional
        The gyromagnetic ratio in MHz/Tesla (default value is for is for 1H).
    field_strengh : float, optional
        MRI field strength (in Tesla).

    Returns
    -------
    relative_frequencies_Hz : array
        A numpy array containing the relative frequencies in Hz for the fat
        peaks.
    fat_rel_amps : array
        The relative amplitude of the fat peaks (the sum of all amplitudes
        is normalized to 1).

    """

    # relative amplitudes of the peaks
    fat_rel_amps = np.asarray([1.0])
    fat_rel_amps /= fat_rel_amps.sum()

    freqs_3T_1H = np.asarray([420])
    relative_frequencies_Hz = gamma / 42.58 * field_strength / 3 * freqs_3T_1H

    return -relative_frequencies_Hz, fat_rel_amps


def fat_model(field_strength, num_peaks=6, gamma=42.58):
    """Simple, single peak fat model.

    Parameters
    ----------
    field_strengh : float, optional
        MRI field strength (in Tesla).
    num_peaks : int
        The number of fat peaks included in the model. Must be in the range
        ``[1, 6]``.
    gamma : float, optional
        The gyromagnetic ratio in MHz/Tesla (default value is for is for 1H).

    Returns
    -------
    relative_frequencies_Hz : array
        A numpy array containing the relative frequencies in Hz for the fat
        peaks.
    fat_rel_amps : array
        The relative amplitude of the fat peaks (the sum of all amplitudes
        is normalized to 1).

    """
    if num_peaks == 1:
        return single_peak_fat_model(gamma=gamma, field_strength=field_strength)
    elif num_peaks == 3:
        return three_peak_fat_model(gamma=gamma, field_strength=field_strength)
    elif num_peaks <= 6:
        return six_peak_fat_model(
            gamma=gamma, field_strength=field_strength, components=num_peaks
        )
    else:
        raise ValueError("unsupported number of peaks")



from cupy.cuda import device
import matplotlib.pyplot as plt

from mrrt.mri.operators.bench.bench_mri import (
    bench_mri_2d_nocoils_nofieldmap,
    bench_mri_2d_16coils_nofieldmap,
    bench_mri_2d_16coils_fieldmap,
    bench_mri_2d_16coils_fieldmap_multispectral,
    bench_mri_3d_nocoils_nofieldmap,
    bench_mri_3d_nocoils_fieldmap,
    bench_mri_3d_16coils_nofieldmap,
    bench_mri_3d_16coils_fieldmap,
)


def run_all_bench(
    plot_timings=True,
    include_GPU=True,
    include_CPU=True,
    print_timings=True,
    phasings=["real", "complex"],
    save_dir="/tmp",
):
    common_kwargs = dict(
        plot_timings=plot_timings,
        include_GPU=include_GPU,
        include_CPU=include_CPU,
        print_timings=print_timings,
        phasings=phasings,
        save_dir=save_dir,
    )
    # 2d cases
    bench_mri_2d_nocoils_nofieldmap(navg_time=10, **common_kwargs)
    bench_mri_2d_16coils_nofieldmap(navg_time=4, **common_kwargs)
    bench_mri_2d_16coils_fieldmap(navg_time=4, **common_kwargs)
    bench_mri_2d_16coils_fieldmap_multispectral(
        navg_time=2, nspectra=2, **common_kwargs
    )

    # 3d cases
    bench_mri_3d_nocoils_nofieldmap(navg_time=4, **common_kwargs)
    bench_mri_3d_nocoils_fieldmap(navg_time=4, **common_kwargs)
    bench_mri_3d_16coils_nofieldmap(navg_time=2, **common_kwargs)
    bench_mri_3d_16coils_fieldmap(navg_time=2, **common_kwargs)


common_kwargs = dict(
    plot_timings=True,
    include_GPU=True,
    include_CPU=True,
    print_timings=True,
    phasings=["complex"],
    save_dir="/tmp",
)


device.get_cusparse_handle()

# cupy.fft.cache.enable()

# bench_mri_3d_16coils_fieldmap(navg_time=2, **common_kwargs)
# bench_mri_2d_16coils_nofieldmap(navg_time=64, **common_kwargs)
bench_mri_2d_16coils_fieldmap(navg_time=32, **common_kwargs)
# bench_mri_2d_16coils_fieldmap_multispectral(
#     navg_time=64, nspectra=2, **common_kwargs
# )
plt.show()



from functools import partial
from math import sqrt
import warnings

import numpy as np

from mrrt.operators import LinearOperatorMulti
from mrrt.utils import config, embed, fftn, ifftn, masker, prod, profile
from mrrt.utils import complexify as as_complex_array
from mrrt.mri.operators._pixel_basis import PixelBasis

if config.have_pyfftw:
    from mrrt.utils import build_fftn, build_ifftn

if config.have_cupy:
    import cupy


class MRI_Cartesian(LinearOperatorMulti):
    """ n-dimensional Fourier Transform operator with optional sampling mask.
    """

    def __init__(
        self,
        arr_shape,
        order="F",
        arr_dtype=np.float32,
        use_fft_shifts=True,
        sample_mask=None,
        ortho=False,
        coil_sensitivities=None,
        force_real_image=False,
        debug=False,
        preplan_pyfftw=True,
        pyfftw_threads=None,
        fft_axes=None,
        fftshift_axes=None,
        planner_effort="FFTW_ESTIMATE",
        loop_over_coils=False,
        preserve_memory=False,
        disable_warnings=False,
        im_mask=None,
        pixel_basis="dirac",
        rel_fov=None,
        **kwargs,
    ):
        """Cartesian MRI Operator  (with partial FFT and coil maps).

        Parameters
        ----------
        arr_shape : int
            shape of the array
        order : {'C','F'}, optional
            array ordering that will be assumed if inputs/outputs need to be
            reshaped
        arr_dtype : numpy.dtype, optional
            dtype for the array
        sample_mask : array_like, optional
            boolean mask of which FFT coefficients to keep
        coil_sensitivities : array, optional
            Array of coil sensitivities.
        ortho : bool, optional
            if True, change the normalizeation to the orthogonal case
        preplan_pyfftw : bool, optional
            if True, precompute the pyFFTW plan upon object creation
        pyfftw_threads : int, optional
            number of threads to be used by pyFFTW.  defaults to
            multiprocessing.cpu_count() // 2.
        use_fft_shifts : bool, optional
            If False, do not apply any FFT shifts
        fft_axes : tuple or None, optional
            Specify a subset of the axes to transform.  The default is to
            transform all axes.
        fftshift_axes : tuple or None, optional
            Specify a subset of the axes to fftshift.  The default is to
            shift all axes.
        im_mask : ndarray or None, optional
            Image domain mask
        force_real_image : bool, optional
        loop_over_coils : bool, optional
            If True, memory required is lower, but speed will be slower.
        preserve_memory : bool, optional
            If False, conjugate copy of coils won't be precomputed
        debug : bool, optional

        Additional Parameters
        ---------------------
        nd_input : bool, optional
        nd_output : bool, optional

        """
        if isinstance(arr_shape, (np.ndarray, list)):
            # retrieve shape from array
            arr_shape = tuple(arr_shape)
        if not isinstance(arr_shape, tuple):
            raise ValueError("expected array_shape to be a tuple or list")

        self.arr_shape = arr_shape
        self.ndim = len(arr_shape)
        self.order = order
        self.use_fft_shifts = use_fft_shifts
        self.disable_warnings = disable_warnings

        if "loc_in" in kwargs and kwargs["loc_in"] == "gpu":
            xp = cupy
            on_gpu = True
        else:
            xp = np
            on_gpu = False
        self._on_gpu = on_gpu

        if sample_mask is not None:
            # masking faster if continguity of mask matches
            if self.order == "F":
                sample_mask = xp.asfortranarray(sample_mask)
            elif self.order == "C":
                sample_mask = xp.ascontiguousarray(sample_mask)
            else:
                raise ValueError("order must be C or F")
        self.sample_mask = sample_mask

        self.force_real_image = force_real_image
        self.debug = debug
        if self.sample_mask is not None:
            if sample_mask.shape != arr_shape:
                raise ValueError("sample mask shape must match arr_shape")
            # make sure it is boolean
            self.sample_mask = self.sample_mask > 0
            # # prestore raveled mask indices to save time later during masking
            # self.sample_mask_idx = xp.where(
            #     self.sample_mask.ravel(order=self.order)
            # )

        self.preserve_memory = preserve_memory
        self.loop_over_coils = loop_over_coils

        # can specify a subset of the axes to perform the FFT/FFTshifts over
        self.fft_axes = fft_axes
        if self.fft_axes is None:
            if self.order == "C":
                # last ndim axes
                self.fft_axes = tuple([-ax for ax in range(self.ndim, 0, -1)])
            else:
                # first ndim axes
                self.fft_axes = tuple(range(self.ndim))

        if fftshift_axes is None:
            self.fftshift_axes = self.fft_axes
        else:
            self.fftshift_axes = fftshift_axes

        # configure scaling  (e.g. unitary operator or not)
        self.ortho = ortho
        if self.fft_axes is None:
            Ntrans = prod(self.arr_shape)
        else:
            Ntrans = prod(np.asarray(self.arr_shape)[np.asarray(self.fft_axes)])
        if self.ortho:
            # sqrt of product of shape along axes where FFT is performed
            self.scale_ortho = sqrt(Ntrans)
            self.gpu_scale_inverse = 1  # self.scale_ortho / Ntrans
            # self.gpu_scale_forward = self.scale_ortho
        else:
            self.scale_ortho = None
            self.gpu_scale_inverse = 1 / Ntrans

        if "mask_out" in kwargs:
            raise ValueError(
                "This operator specifies `mask_out` via the "
                "parameter `sample_mask"
            )

        if ("mask_in" in kwargs) or ("mask_out" in kwargs):
            raise ValueError(
                "This operator specifies `mask_in` via the "
                "parameter `im_mask"
            )

        if coil_sensitivities is not None:
            if coil_sensitivities.ndim == self.ndim + 1:
                # self.arr_shape + (Ncoils, )
                if self.order == "C":
                    Ncoils = coil_sensitivities.shape[0]
                else:
                    Ncoils = coil_sensitivities.shape[-1]
                Nmaps = 1
            elif coil_sensitivities.ndim == self.ndim + 2:
                # case with multiple maps (e.g. ESPIRIT soft-SENSE)
                # self.arr_shape + (Ncoils, Nmaps)
                if self.order == "C":
                    Ncoils = coil_sensitivities.shape[1]
                    Nmaps = coil_sensitivities.shape[0]
                else:
                    Ncoils = coil_sensitivities.shape[-2]
                    Nmaps = coil_sensitivities.shape[-1]
            else:
                # determine based on size
                Ncoils = coil_sensitivities.size / prod(self.arr_shape)
            if (Ncoils % 1) != 0:
                raise ValueError("sensitivity map size mismatch")
            self.Ncoils = int(Ncoils)
            self.Nmaps = Nmaps
            if self.order == "C":
                cmap_shape = (self.Nmaps, self.Ncoils) + self.arr_shape
                if not coil_sensitivities.flags.c_contiguous:
                    msg = (
                        "Converting coil_sensitivities to be C contiguous"
                        " (requires a copy).  To avoid the copy, convert to "
                        "Fortran contiguous order prior to calling "
                        "MRI_Cartesian (see np.ascontiguousarray)"
                    )
                    if not self.disable_warnings:
                        warnings.warn(msg)
                    coil_sensitivities = xp.ascontiguousarray(
                        coil_sensitivities
                    )
            else:
                cmap_shape = self.arr_shape + (self.Ncoils, self.Nmaps)
                if not coil_sensitivities.flags.f_contiguous:
                    msg = (
                        "Converting coil_sensitivities to be Fortan contiguous"
                        " (requires a copy).  To avoid the copy, convert to "
                        "Fortran contiguous order prior to calling "
                        "MRI_Cartesian (see np.asfortranarray)"
                    )
                    if not self.disable_warnings:
                        warnings.warn(msg)
                    coil_sensitivities = xp.asfortranarray(coil_sensitivities)
            if tuple(coil_sensitivities.shape) != tuple(cmap_shape):
                coil_sensitivities = coil_sensitivities.reshape(
                    cmap_shape, order=self.order
                )
            if not self.preserve_memory:
                self.coil_sensitivities_conj = xp.conj(coil_sensitivities)
        else:
            self.Ncoils = 1
            self.Nmaps = 1

        if self.Ncoils == 1:
            # TODO: currently has a shape bug if Ncoils == 1
            #       and loop_over_coils = False.
            self.loop_over_coils = True

        self.coil_sensitivities = coil_sensitivities

        if im_mask is not None:
            if im_mask.shape != arr_shape:
                raise ValueError("im_mask shape mismatch")
            if order != "F":
                raise ValueError("only order='F' supported for im_mask case")
            nargin = xp.count_nonzero(im_mask)
            self.im_mask = im_mask
        else:
            nargin = prod(arr_shape)
            self.im_mask = None
        nargin *= self.Nmaps

        # nargout = # of k-space samples
        if sample_mask is not None:
            nargout = xp.count_nonzero(sample_mask) * self.Ncoils
        else:
            nargout = nargin // self.Nmaps * self.Ncoils
        nargout = int(nargout)

        self.idx_orig = None
        self.idx_conj = None
        self.sample_mask_conj = None

        # output of FFTs will be complex, regardless of input type
        self.result_dtype = np.result_type(arr_dtype, np.complex64)

        matvec_allows_repetitions = kwargs.pop(
            "matvec_allows_repetitions", True
        )
        squeeze_reps = kwargs.pop("squeeze_reps", True)
        nd_input = kwargs.pop("nd_input", False)
        nd_output = kwargs.pop("nd_output", False)

        if (self.sample_mask is not None) and nd_output:
            raise ValueError("cannot have both nd_output and sample_mask")
        if nd_output:
            if self.Ncoils == 1:
                shape_out = self.arr_shape
            else:
                if Nmaps == 1:
                    if self.order == "C":
                        shape_out = (self.Ncoils,) + self.arr_shape
                    else:
                        shape_out = self.arr_shape + (self.Ncoils,)
                else:
                    if self.order == "C":
                        shape_out = (self.Nmaps, self.Ncoils) + self.arr_shape
                    else:
                        shape_out = self.arr_shape + (self.Ncoils, self.Nmaps)
        else:
            shape_out = (nargout, 1)

        if self.Nmaps == 1:
            shape_in = self.arr_shape
        else:
            if self.order == "C":
                shape_in = (self.Nmaps,) + self.arr_shape
            else:
                shape_in = self.arr_shape + (self.Nmaps,)
        if self.order == "C":
            self.shape_inM = (self.Nmaps,) + self.arr_shape
        else:
            self.shape_inM = self.arr_shape + (self.Nmaps,)
        self.shape_in1 = self.arr_shape
        self.have_pyfftw = config.have_pyfftw

        if self.on_gpu:
            self.preplan_pyfftw = False
        else:
            self.preplan_pyfftw = preplan_pyfftw if self.have_pyfftw else False

            if self.preplan_pyfftw:
                self._preplan_fft(pyfftw_threads, planner_effort)
                # raise ValueError("Implementation Incomplete")

        if self.on_gpu:
            self.fftn = partial(fftn, xp=cupy)
            self.ifftn = partial(ifftn, xp=cupy)
        else:
            if self.preplan_pyfftw:
                self._preplan_fft(pyfftw_threads, planner_effort)
            else:
                self.fftn = partial(fftn, xp=np)
                self.ifftn = partial(ifftn, xp=np)
        self.fftshift = xp.fft.fftshift
        self.ifftshift = xp.fft.ifftshift

        self.rel_fov = rel_fov

        self.mask = None  # TODO: implement or remove (expected by CUDA code)

        matvec = self.forward
        matvec_adj = self.adjoint
        self.norm_available = False
        self.norm = self._norm

        super(MRI_Cartesian, self).__init__(
            nargin=nargin,
            nargout=nargout,
            matvec=matvec,
            matvec_transp=matvec_adj,
            matvec_adj=matvec_adj,
            nd_input=nd_input or (im_mask is not None),
            nd_output=nd_output,
            shape_in=shape_in,
            shape_out=shape_out,
            order=self.order,
            matvec_allows_repetitions=matvec_allows_repetitions,
            squeeze_reps=squeeze_reps,
            mask_in=im_mask,
            mask_out=None,  # mask_out,
            symmetric=False,  # TODO: set properly
            hermetian=False,  # TODO: set properly
            dtype=self.result_dtype,
            **kwargs,
        )

        self._init_pixel_basis(pixel_basis=pixel_basis)

    def _preplan_fft(self, pyfftw_threads=None, planner_effort="FFTW_MEASURE"):
        """ Use FFTW builders to pre-plan the FFT for faster repeated
        calls. """
        if pyfftw_threads is None:
            import multiprocessing

            pyfftw_threads = max(1, multiprocessing.cpu_count())
        self.pyfftw_threads = pyfftw_threads
        if self.loop_over_coils:
            a_b = np.empty(self.arr_shape, dtype=self.result_dtype)
        else:
            if self.order == "C":
                a_b = np.empty(
                    (self.Ncoils,) + self.arr_shape, dtype=self.result_dtype
                )
            else:
                a_b = np.empty(
                    self.arr_shape + (self.Ncoils,), dtype=self.result_dtype
                )
        self.fftn = build_fftn(
            a_b,
            axes=self.fft_axes,
            threads=pyfftw_threads,
            overwrite_input=False,
            planner_effort=planner_effort,
        )
        self.ifftn = build_ifftn(
            a_b,
            axes=self.fft_axes,
            threads=pyfftw_threads,
            overwrite_input=False,
            planner_effort=planner_effort,
        )
        del a_b

    def _rescale_basis(self):
        if self.basis.transform is not None:
            from mrrt.utils import power_method

            lam = power_method(self.H * self)
            self.basis.transform /= sqrt(lam)
        return self.basis

    def _init_pixel_basis(self, pixel_basis=None, rel_fov=None):
        xp = self.xp
        if pixel_basis is None:
            pixel_basis = self.pixel_basis
        else:
            self.pixel_basis = pixel_basis

        if rel_fov is None:
            rel_fov = self.rel_fov
        else:
            self.rel_fov = rel_fov

        if pixel_basis != "dirac":
            if rel_fov is None:
                raise ValueError(
                    "For pixel_basis = {}, rel_fov must be specfied.".format(
                        pixel_basis
                    )
                )
        else:

            class DummyBasis(object):
                transform = None

            self.basis = DummyBasis()
            return self.basis

        rel_dx = np.asarray(self.arr_shape) / np.asarray(self.rel_fov)
        rel_dx /= rel_dx.max()
        ranges = [
            (xp.arange(s) / s - 0.5) * rel_dx[n]
            for n, s in enumerate(self.arr_shape)
        ]
        ksp = xp.meshgrid(*ranges, indexing="ij", sparse=False)
        ksp = xp.stack([k.ravel(order="F") for k in ksp], axis=-1)

        self.basis = PixelBasis(
            ksp,
            pixel_basis=self.pixel_basis,
            # don't change values below here.  already normalized the
            # k-space properly above
            dx=1.0,
            fov=self.arr_shape,
            Nd=self.arr_shape,
            xp=self.xp,
        )

        self.basis.transform = xp.asarray(
            self.basis.transform, dtype=self.result_dtype
        )
        self.basis.transform = self.basis.transform.reshape(
            self.arr_shape, order="F"
        )

        # rescale basis.transform to maintain a norm-preserving operator
        self.basis = self._rescale_basis()
        return self.basis

    @profile
    def _adjoint_single_rep(self, y, i_map=0):
        use_smaps = self.coil_sensitivities is not None
        xp = self.xp
        if self.loop_over_coils:
            if use_smaps:
                x = xp.zeros(
                    self.shape_in1,
                    order=self.order,
                    dtype=xp.result_type(y, np.complex64),
                )
            for coil in range(self.Ncoils):
                if self.order == "C":
                    coil_slice = (coil, Ellipsis)
                    imap_coil_slice = (i_map, coil, Ellipsis)
                else:
                    coil_slice = (Ellipsis, coil)
                    imap_coil_slice = (Ellipsis, coil, i_map)
                y0 = y[coil_slice]
                if self.basis.transform is not None:
                    y0 *= xp.conj(self.basis.transform)
                if self.use_fft_shifts:
                    y0 = self.ifftshift(y0, axes=self.fftshift_axes)
                if self.preplan_pyfftw:
                    x0 = self.ifftn(y0)
                else:
                    x0 = self.ifftn(y0, axes=self.fft_axes)
                if self.use_fft_shifts:
                    x0 = self.fftshift(x0, axes=self.fftshift_axes)
                if use_smaps:
                    if self.preserve_memory:
                        x += x0 * xp.conj(
                            self.coil_sensitivities[imap_coil_slice]
                        )
                    else:
                        x += x0 * self.coil_sensitivities_conj[imap_coil_slice]
                else:
                    x = x0
        else:
            if self.use_fft_shifts:
                y = self.ifftshift(y, axes=self.fftshift_axes)
            if self.basis.transform is not None:
                if self.order == "C":
                    y *= xp.conj(self.basis.transform[np.newaxis, ...])
                else:
                    y *= xp.conj(self.basis.transform[..., np.newaxis])
            if self.preplan_pyfftw:
                x = self.ifftn(y)
            else:
                x = self.ifftn(y, axes=self.fft_axes)
            if self.use_fft_shifts:
                x = self.fftshift(x, axes=self.fftshift_axes)
            if use_smaps:
                if self.order == "C":
                    imap_slice = (i_map, Ellipsis)
                    coil_axis = 0
                else:
                    imap_slice = (Ellipsis, i_map)
                    coil_axis = -1
                if self.preserve_memory:
                    x *= xp.conj(self.coil_sensitivities[imap_slice])
                else:
                    x *= self.coil_sensitivities_conj[imap_slice]
                x = x.sum(coil_axis)  # sum over coils
        return x

    @profile
    def adjoint(self, y):
        # TODO: add test for this case and the coils + sample_mask case
        # if y.ndim == 1 or y.shape[-1] == 1:
        xp = self.xp
        ncoils = self.Ncoils
        nreps = int(y.size / self.shape[0])
        if self.sample_mask is not None:
            if y.ndim == 1 and self.ndim > 1:
                y = y[:, np.newaxis]
            nmask = xp.count_nonzero(self.sample_mask)
            if self.on_gpu:
                nmask = nmask.get()
            if y.shape[0] != nmask:
                if self.order == "C":
                    y = y.reshape((-1, nmask), order=self.order)
                else:
                    y = y.reshape((nmask, -1), order=self.order)
            y = embed(y, mask=self.sample_mask, order=self.order)
        if nreps == 1:
            # 1D or single repetition or single coil nD
            if self.order == "C":
                y = y.reshape((ncoils,) + self.shape_in1, order=self.order)
            else:
                y = y.reshape(self.shape_in1 + (ncoils,), order=self.order)
            if self.Nmaps == 1:
                x = self._adjoint_single_rep(y, i_map=0)
            else:
                x = xp.zeros(
                    self.shape_inM, dtype=xp.result_type(y, xp.complex64)
                )
                if self.order == "C":
                    for i_map in range(self.Nmaps):
                        x[i_map, ...] = self._adjoint_single_rep(y, i_map=i_map)
                else:
                    for i_map in range(self.Nmaps):
                        x[..., i_map] = self._adjoint_single_rep(y, i_map=i_map)
        else:
            if self.order == "C":
                y = y.reshape(
                    (nreps, ncoils) + self.shape_in1, order=self.order
                )  # or shape_out?
                x_shape = (nreps,) + self.shape_inM
            else:
                y = y.reshape(
                    self.shape_in1 + (ncoils, nreps), order=self.order
                )  # or shape_out?
                x_shape = self.shape_inM + (nreps,)

            x = xp.zeros(x_shape, dtype=xp.result_type(y, xp.complex64))
            for i_map in range(self.Nmaps):
                if self.order == "C":
                    for rep in range(nreps):
                        x[rep, i_map, ...] = self._adjoint_single_rep(
                            y[rep, ...], i_map=i_map
                        )
                else:
                    for rep in range(nreps):
                        x[..., i_map, rep] = self._adjoint_single_rep(
                            y[..., rep], i_map=i_map
                        )
        #        if self.im_mask:
        #            x = masker(x, self.im_mask, order=self.order)
        if self.ortho:
            x *= self.scale_ortho
        if x.dtype != self.result_dtype:
            x = x.astype(self.result_dtype)
        if self.force_real_image:
            # x = x.real.astype(self.result_dtype)
            if x.dtype in [np.complex64, np.complex128]:
                x.imag[:] = 0
        return x

    @profile
    def _forward_single_rep(self, x, i_map=0):
        xp = self.xp
        if self.order == "C":
            y_shape = (self.Ncoils,) + self.shape_in1
        else:
            y_shape = self.shape_in1 + (self.Ncoils,)
        y = xp.zeros(
            y_shape, dtype=xp.result_type(x, xp.complex64), order=self.order
        )
        use_smaps = self.coil_sensitivities is not None
        if self.loop_over_coils:
            for coil in range(self.Ncoils):
                if use_smaps:
                    if self.order == "C":
                        xc = x * self.coil_sensitivities[i_map, coil, ...]
                    else:
                        xc = x * self.coil_sensitivities[..., coil, i_map]
                else:
                    xc = x
                if self.order == "C":
                    coil_slice = (coil, Ellipsis)
                else:
                    coil_slice = (Ellipsis, coil)
                if self.debug:
                    print("x.shape = {}".format(x.shape))
                    print("xc.shape = {}".format(xc.shape))
                    print("y.shape = {}".format(y.shape))
                if self.use_fft_shifts:
                    y[coil_slice] = self.ifftshift(xc, axes=self.fftshift_axes)
                else:
                    y[coil_slice] = xc
                if self.preplan_pyfftw:
                    y[coil_slice] = self.fftn(y[coil_slice])
                else:
                    y[coil_slice] = self.fftn(y[coil_slice], axes=self.fft_axes)
                if self.use_fft_shifts:
                    y[coil_slice] = self.fftshift(
                        y[coil_slice], axes=self.fftshift_axes
                    )
                if self.basis.transform is not None:
                    y *= self.basis.transform
        else:
            if use_smaps:
                if self.order == "C":
                    x = as_complex_array(x)[xp.newaxis, ...]
                    x = x * self.coil_sensitivities[i_map, ...]
                else:
                    x = as_complex_array(x)[..., xp.newaxis]
                    x = x * self.coil_sensitivities[..., i_map]
            if self.use_fft_shifts:
                x = self.ifftshift(x, axes=self.fftshift_axes)
            if self.preplan_pyfftw:
                y = self.fftn(x)
            else:
                y = self.fftn(x, axes=self.fft_axes)
            if self.use_fft_shifts:
                y = self.fftshift(y, axes=self.fftshift_axes)
            if self.basis.transform is not None:
                if self.order == "C":
                    y *= self.basis.transform[xp.newaxis, ...]
                else:
                    y *= self.basis.transform[..., xp.newaxis]

        if self.sample_mask is not None:
            # y = y[self.sample_mask]
            y = masker(
                y,
                mask=self.sample_mask,
                order=self.order,
                # mask_idx_ravel=self.sample_mask_idx,
            )
        return y

    @profile
    def forward(self, x):
        xp = self.xp
        if self.force_real_image:
            if x.dtype in [np.complex64, np.complex128]:
                x.imag[:] = 0
        if self.im_mask is None:
            size_1rep = self.nargin
        else:
            size_1rep = prod(self.shape_inM)
        if x.size < size_1rep:
            raise ValueError("data, x, too small to transform.")
        elif x.size == size_1rep:
            nreps = 1
            # 1D or single repetition nD
            x = x.reshape(self.shape_inM, order=self.order)
            if self.order == "C":
                y = self._forward_single_rep(x[0, ...], i_map=0)
                for i_map in range(1, self.Nmaps):
                    y += self._forward_single_rep(x[i_map, ...], i_map=i_map)
            else:
                y = self._forward_single_rep(x[..., 0], i_map=0)
                for i_map in range(1, self.Nmaps):
                    y += self._forward_single_rep(x[..., i_map], i_map=i_map)
        else:
            if self.order == "C":
                nreps = x.shape[0]
                x_shape = (nreps,) + self.shape_inM
                if self.sample_mask is not None:
                    y_shape = (nreps, self.Ncoils, self.nargout)
                else:
                    y_shape = (nreps, self.Ncoils) + self.shape_in1
            else:
                nreps = x.shape[-1]
                x_shape = self.shape_inM + (nreps,)
                if self.sample_mask is not None:
                    y_shape = (self.nargout, self.Ncoils, nreps)
                else:
                    y_shape = self.shape_in1 + (self.Ncoils, nreps)
            x = x.reshape(x_shape, order=self.order)
            if self.sample_mask is not None:
                # number of samples by number of repetitions
                y = xp.zeros(
                    y_shape,
                    dtype=xp.result_type(x, xp.complex64),
                    order=self.order,
                )
            else:
                y = xp.zeros(
                    y_shape,
                    dtype=xp.result_type(x, xp.complex64),
                    order=self.order,
                )
            if self.order == "C":
                for rep in range(nreps):
                    y[rep, ...] = self._forward_single_rep(
                        x[rep, 0, ...], i_map=0
                    ).reshape(y.shape[1:], order=self.order)
                    for i_map in range(1, self.Nmaps):
                        y[rep, ...] += self._forward_single_rep(
                            x[rep, i_map, ...], i_map=i_map
                        ).reshape(y.shape[1:], order=self.order)
            else:
                for rep in range(nreps):
                    y[..., rep] = self._forward_single_rep(
                        x[..., 0, rep], i_map=0
                    ).reshape(y.shape[:-1], order=self.order)
                    for i_map in range(1, self.Nmaps):
                        y[..., rep] += self._forward_single_rep(
                            x[..., i_map, rep], i_map=i_map
                        ).reshape(y.shape[:-1], order=self.order)
        if self.ortho:
            y /= self.scale_ortho

        if y.dtype != self.result_dtype:
            y = y.astype(self.result_dtype)

        return y

    def _norm_single_rep(self, x):
        # chain forward and adjoint together for a single repetition
        y = self._forward_single_rep(x)
        if self.sample_mask is not None:
            y = embed(y, mask=self.sample_mask, order=self.order)
        if self.order == "C":
            y_shape = (self.Ncoils,) + self.shape_in1
        else:
            y_shape = self.shape_in1 + (self.Ncoils,)
        y = y.reshape(y_shape, order=self.order)
        x = self._adjoint_single_rep(y)
        return x

    @profile
    def _norm(self, x):
        # forward transform, immediately followed by inverse transform
        # slightly faster than calling self.adjoint(self.forward(x))
        xp = self.xp
        if self.force_real_image:
            if x.dtype in [np.complex64, np.complex128]:
                x.imag[:] = 0
        if self.im_mask is None:
            size_1rep = self.nargin
        else:
            size_1rep = prod(self.shape_inM)
        if x.size < size_1rep:
            raise ValueError("data, x, too small to transform.")
        elif x.size == size_1rep:
            nreps = 1
            # 1D or single repetition nD
            x = x.reshape(self.shape_inM, order=self.order)
            y = self._norm_single_rep(x)
        else:
            nreps = x.size // size_1rep
            if self.order == "C":
                x = x.reshape((nreps,) + self.shape_inM, order=self.order)
                y = xp.zeros_like(x)
                for rep in range(nreps):
                    y[rep, ...] = self._norm_single_rep(x[rep, ...]).reshape(
                        y.shape[1:], order=self.order
                    )
            else:
                x = x.reshape(self.shape_inM + (nreps,), order=self.order)
                y = xp.zeros_like(x)
                for rep in range(nreps):
                    y[..., rep] = self._norm_single_rep(x[..., rep]).reshape(
                        y.shape[:-1], order=self.order
                    )
        if y.dtype != self.result_dtype:
            y = y.astype(self.result_dtype)

        if not self.nd_input:
            if self.squeeze_reps_in and (nreps == 1):
                y = xp.ravel(y, order=self.order)
            else:
                if self.order == "C":
                    y = y.reshape((nreps, -1), order=self.order)
                else:
                    y = y.reshape((-1, nreps), order=self.order)
        return y



from itertools import product
import numpy as np
import pytest

from mrrt.mri import sensemap_sim

from mrrt.utils import config

all_xp = [np]
if config.have_cupy:
    import cupy

    all_xp += [cupy]


@pytest.mark.parametrize(
    "xp, dtype", product(all_xp, [np.complex64, np.complex128])
)
def test_sensemap_sim_2d(xp, dtype):
    shape = (64, 64)
    spacings = tuple(240 / s for s in shape)
    ncoil = 8
    smap = sensemap_sim(shape=shape, spacings=spacings, ncoil=8, dtype=dtype)
    assert smap.shape == shape + (ncoil,)
    assert smap.dtype == dtype



"""System matrix objects (operators) for Cartesian and Non-Cartesian MRI.

"""
from numpy.testing import Tester

# Cartesian MRI Operator (includes multi-coil and fieldmap processing)
from ._MRI_Cartesian import MRI_Cartesian  # noqa

# NonUniform FFT operator
try:
    # Basic NUFFT Operator
    from ._NUFFT import NUFFT_Operator  # noqa

    # MRI Operator (includes multi-coil and fieldmap processing)
    from ._MRI import MRI_Operator  # noqa
except (ImportError, NameError):
    print(
        "NUFFT_Operator & MRI_OPerator not available (requires mrrt.nufft "
        "and mrrt.operators packages)"
    )

test = Tester().test



# import cupy
from mrrt.mri.operators.bench.bench_mri import bench_mri_3d_16coils_fieldmap

# if hasattr(cupy.fft, "cache") and hasattr(cupy.fft.cache, "enable"):
#     cupy.fft.cache.enable()

bench_mri_3d_16coils_fieldmap(
    plot_timings=True, include_CPU=False, include_GPU=True
)



"""Coil compression, pre-whitening and sensitivy map estimation utilities.

Much of this code is modified from idmrmrd-python-tools
https://github.com/ismrmrd/ismrmrd-python-tools

Much code that was 2D only has been updated to work with 3D (or nD) data.
The user can specify which axis of the array corresponds to the coils.

The function calculate_csm_inati_iter is an adaptation of a coil sensitivity
map estimation function by Souheil Inati that is included in Gadgetron.

"""

from ._coil_maps import *  # noqa
from ._coil_pca import *  # noqa
from ._prewhiten import *  # noqa
from ._bias import *  # noqa



from os.path import join as pjoin

# Format expected by setup.py and doc/source/conf.py: string of form "X.Y.Z"
_version_major = 0
_version_minor = 3
_version_micro = 4  # use "" for first of series, number for 1 and above
_version_extra = ""  # use "dev0" for developemnt, "" for full release

# Construct full version string from these.
_ver = [_version_major, _version_minor]
if _version_micro:
    _ver.append(_version_micro)
if _version_extra:
    _ver.append(_version_extra)

__version__ = ".".join(map(str, _ver))

CLASSIFIERS = [
    "Development Status :: 4 - Beta",
    "Environment :: Console",
    "Intended Audience :: Science/Research",
    "License :: OSI Approved :: BSD License",
    "Operating System :: OS Independent",
    "Programming Language :: Python",
    "Programming Language :: Python :: 3",
    "Programming Language :: Python :: 3.6",
    "Programming Language :: Python :: 3.7",
    "Programming Language :: Python :: 3.8",
    "Programming Language :: Python :: 3 :: Only",
    "Programming Language :: Python :: Implementation :: CPython",
    "Topic :: Scientific/Engineering",
]

# Description should be a one-liner:
description = "MRI Reconstruction Tools"
# Long description will go up on the pypi page
long_description = """

mrrt.mri
==========
MRI Reconstruction Tools

This module contains linear operators for Cartesian and Non-Cartesian MRI as
well as a number of utilities for coil sensitivity map estimation,
coil compression, kspace data prewhitening, phantoms and field map
approximation.

.. _README: https://github.com/mritools/mrrt.mri/blob/master/README.md

License
=======
``mrrt.mri`` is licensed under the terms of the BSD 3-clause license. See
the file "LICENSE.txt" and "LICENSES_bundled.txt" for information on the
history of this software, terms & conditions for usage, and a DISCLAIMER OF ALL
WARRANTIES.

All trademarks referenced herein are property of their respective holders.

Copyright (c) 2014-2020,
Gregory R. Lee, Cincinnati Children's Hospital Medical Center.
"""

NAME = "mrrt.mri"
MAINTAINER = "Gregory R. Lee"
MAINTAINER_EMAIL = "grlee77@gmail.com"
DESCRIPTION = description
LONG_DESCRIPTION = long_description
URL = "http://github.com/mritools/mrrt.mri"
DOWNLOAD_URL = ""
LICENSE = "BSD"
AUTHOR = "Gregory R. Lee"
AUTHOR_EMAIL = "grlee77@gmail.com"
PLATFORMS = "OS Independent"
MAJOR = _version_major
MINOR = _version_minor
MICRO = _version_micro
VERSION = __version__
PACKAGE_DATA = {
    "mrrt.mri": [
        pjoin("tests", "*"),
        pjoin("data", "*"),
        pjoin("coils", "tests", "*"),
        pjoin("field_maps", "tests", "*"),
        pjoin("operators", "tests", "*"),
        pjoin("sim", "tests", "*"),
    ]
}
REQUIRES = ["numpy", "mrrt.utils", "mrrt.nufft", "mrrt.operators"]
PYTHON_REQUIRES = ">= 3.6"



import numpy as np
from mrrt.utils import embed, fftnc, get_array_module, ifftnc


def mri_partial_fourier_nd(
    partial_kspace,
    pf_mask,
    niter=5,
    tw_inner=8,
    tw_outer=3,
    fill_conj=False,
    init=None,
    verbose=False,
    show=False,
    return_all_estimates=False,
    xp=None,
):
    """Partial Fourier reconstruction.

    Parameters
    ----------

    Returns
    -------

    Notes
    -----
    The implementation is based on a multi-dimensional iterative reconstruction
    technique as described in [1]_.  This is an extension of the 1D iterative
    methods described in [2]_ and [3]_.  The concept of partial Fourier imaging
    was first proposed in [4]_, [5]_.

    References
    ----------
    .. [1] Xu, Y. and Haacke, E. M.  Partial Fourier imaging in
        multi-dimensions: A means to save a full factor of two in time.
        J. Magn. Reson. Imaging, 2001; 14:628–635.
        doi:10.1002/jmri.1228

    .. [2] Haacke, E.; Lindskogj, E. & Lin, W. A fast, iterative,
        partial-Fourier technique capable of local phase recovery.
        Journal of Magnetic Resonance, 1991; 92:126-145.

    .. [3] Liang, Z.-P.; Boada, F.; Constable, R. T.; Haacke, E.; Lauterbur,
        P. & Smith, M. Constrained reconstruction methods in MR imaging.
        Rev Magn Reson Med, 1992; 4: 67-185

    .. [4] Margosian, P.; Schmitt, F. & Purdy, D. Faster MR imaging: imaging
        with half the data Health Care Instrum, 1986; 1:195.

    .. [5] Feinberg, D. A.; Hale, J. D.; Watts, J. C.; Kaufman, L. & Mark, A.
        Halving MR imaging time by conjugation: demonstration at 3.5 kG.
        Radiology, 1986, 161, 527-531

    """
    xp, on_gpu = get_array_module(partial_kspace, xp)
    partial_kspace = xp.asarray(partial_kspace)
    # dtype = partial_kspace.dtype

    pf_mask = xp.asarray(pf_mask)
    if pf_mask.dtype != xp.bool:
        pf_mask = pf_mask.astype(xp.bool)
    im_shape = pf_mask.shape
    ndim = pf_mask.ndim

    if not xp.all(xp.asarray(im_shape) % 2 == 0):
        raise ValueError(
            "This function assumes all k-space dimensions have even length."
        )

    if partial_kspace.size != xp.count_nonzero(pf_mask):
        raise ValueError(
            "partial kspace should have total size equal to the number of "
            "non-zeros in pf_mask"
        )

    kspace_init = embed(partial_kspace, pf_mask)
    img_est = ifftnc(kspace_init)

    lr_kspace = xp.zeros_like(kspace_init)

    nz = xp.where(pf_mask)
    lr_slices = [slice(None)] * ndim
    pf_slices = [slice(None)] * ndim
    win2_slices = [slice(None)] * ndim
    lr_shape = [0] * ndim
    # pf_shape = [0, ]*ndim
    win2_shape = [0] * ndim
    for d in range(ndim):
        nz_min = xp.min(nz[d])
        nz_max = xp.max(nz[d])
        if hasattr(nz_min, "get"):
            # 0-dim GPU array to scalar
            nz_min, nz_max = nz_min.get(), nz_max.get()
        i_mid = im_shape[d] // 2
        if nz_min == 0:
            i_end = nz_max
            width = i_end - i_mid
        else:
            i_start = nz_min
            width = i_mid - i_start
        lr_slices[d] = slice(i_mid - width, i_mid + width + 1)
        lr_shape[d] = 2 * width + 1

        # pf_slices[d] = slice(nz_min, nz_max + 1)
        pf_shape = nz_max - nz_min + 1
        pf_slices[d] = slice(nz_min, nz_max + 1)
        win2_shape[d] = pf_shape + tw_outer
        if nz_min == 0:
            # pf_slices[d] = slice(nz_min, nz_max + 1 + tw_outer)
            win2_slices[d] = slice(tw_outer, tw_outer + pf_shape)
        else:
            # pf_slices[d] = slice(nz_min - tw_outer, nz_max + 1)
            win2_slices[d] = slice(pf_shape)

    lr_slices = tuple(lr_slices)
    win2_slices = tuple(win2_slices)
    pf_slices = tuple(pf_slices)
    # lr_mask = xp.zeros(pf_mask, dtype=xp.zeros)
    lr_win = hanning_apodization_window(lr_shape, tw_inner, xp)
    lr_kspace[lr_slices] = kspace_init[lr_slices] * lr_win

    img_lr = ifftnc(lr_kspace)
    phi = xp.angle(img_lr)

    pf_win = hanning_apodization_window(win2_shape, tw_outer, xp)[win2_slices]

    lr_mask = xp.zeros(pf_mask.shape, dtype=xp.float32)
    lr_mask[lr_slices] = lr_win

    win2_mask = xp.zeros(pf_mask.shape, dtype=xp.float32)
    win2_mask[pf_slices] = pf_win

    if show and ndim == 2:
        from matplotlib import pyplot as plt

        fig, axes = plt.subplots(2, 2)
        axes = axes.ravel()
        axes[0].imshow(pf_mask)
        axes[0].set_title("PF mask")
        axes[1].imshow(lr_mask)
        axes[1].set_title("LR Filter")
        axes[2].imshow(win2_mask)
        axes[2].set_title("Filter")
        axes[3].imshow(xp.abs(img_est).T)
        axes[3].set_title("Initial Estimate")
        for ax in axes:
            ax.set_xticklabels("")
            ax.set_yticklabels("")

    if verbose:
        norm0 = xp.linalg.norm(img_est)
        max0 = xp.max(xp.abs(img_est))
    if return_all_estimates:
        all_img_est = [img_est]

    # POCS iterations
    for i in range(niter):
        # step 5
        rho1 = xp.abs(img_est) * xp.exp(1j * phi)

        if verbose:
            change2 = xp.linalg.norm(rho1 - img_est) / norm0
            change1 = xp.max(xp.abs(rho1 - img_est)) / max0
            print("change = {}%% {}%%".format(100 * change2, 100 * change1))

        # step 6
        s1 = fftnc(rho1)

        # steps 7 & 8
        full_kspace = win2_mask * kspace_init + (1 - win2_mask) * s1

        # step 9
        img_est = ifftnc(full_kspace)
        if return_all_estimates:
            all_img_est.append(img_est)

    if return_all_estimates:
        return all_img_est
    return img_est


def hanning_apodization_window(shape, transition_length, xp=np):
    # n-dimensional equivalent of ir_mri_pf_gen_window
    shape = tuple(shape)
    ndim = len(shape)

    if xp.isscalar(transition_length):
        transition_lengths = (transition_length,) * ndim
    else:
        transition_lengths = tuple(transition_length)

    if len(transition_lengths) != ndim:
        raise ValueError(
            "shape and transition_lengths must have the same length"
        )

    awin = 1.0
    for i, tl in enumerate(transition_lengths):
        # (tl + 1) because end points are zero
        hann_len = (tl + 1) * 2 + 1  # always odd so central value will be 1
        hann_1d = xp.hanning(hann_len)[1:-1]  # crop first and last zero
        if hann_1d.max() != 1:
            hann_1d /= hann_1d.max()
        hann_1d_stretch = xp.ones((shape[i],))
        hann_1d_stretch[:tl] = hann_1d[:tl]
        hann_1d_stretch[-tl:] = hann_1d[-tl:]
        s = [1] * ndim
        s[i] = hann_1d_stretch.size
        awin = awin * xp.reshape(hann_1d_stretch, s)  # broadcasting
    awin /= awin.max()
    return awin
